# SPDX-License-Identifier: MIT
import struct

from ..utils import *

from .asc import StandardASC
from .asc.base import *

class PMPMessage(Register64):
    TYPE = 56, 44

class PMP_Startup(PMPMessage):
    TYPE = 56, 44, Constant(0x00)

class PMP_Configure(PMPMessage):
    TYPE = 56, 44, Constant(0x10)
    DVA = 47, 0

class PMP_Configure_Ack(PMPMessage):
    TYPE = 56, 44, Constant(0x20)
    UNK = 47, 0

class PMP_Init1(PMPMessage):
    TYPE = 56, 44, Constant(0x200)
    UNK1 = 43, 16
    UNK2 = 15, 0

class PMP_Init1_Ack(PMPMessage):
    TYPE = 56, 44, Constant(0x201)
    UNK1 = 43, 16
    UNK2 = 15, 0

class PMP_Init2(PMPMessage):
    TYPE = 56, 44, Constant(0x202)
    UNK1 = 43, 16
    UNK2 = 15, 0

class PMP_Init2_Ack(PMPMessage):
    TYPE = 56, 44, Constant(0x203)
    UNK1 = 43, 16
    UNK2 = 15, 0
    
class PMP_Unk(PMPMessage):
    TYPE = 56, 44, Constant(0x100)
    UNK1 = 43, 16
    UNK2 = 15, 0

class PMP_Unk_Ack(PMPMessage):
    TYPE = 56, 44, Constant(0x110)
    UNK1 = 43, 16
    UNK2 = 15, 0

class PMP_DevPwr(PMPMessage):
    TYPE = 56, 44, Constant(0x20e)
    DEV = 31, 16
    STATE = 15, 0

class PMP_DevPwr_Sync(PMPMessage):
    TYPE = 56, 44, Constant(0x208)
    DEV = 31, 16
    STATE = 15, 0

class PMP_DevPwr_Ack(PMPMessage):
    TYPE = 56, 44, Constant(0x209)
    DEV = 31, 16
    STATE = 15, 0

class PMPEndpoint(ASCBaseEndpoint):
    BASE_MESSAGE = PMPMessage
    SHORT = "pmpep"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shmem = self.shmem_dva = None
        self.init_complete = False
        self.init1_acked = False
        self.init2_acked = False
        self.unk_acked = False

    @msg_handler(0x00, PMP_Startup)
    def Startup(self, msg):
        self.log("Starting up")

        self.shmem, self.shmem_dva = self.asc.ioalloc(0x10000)
        
        self.send_init_config()
        return True

    def send_init_config(self):
        self.asc.p.memset32(self.shmem, 0, 0x10000)
        dram_config = self.asc.u.adt["arm-io/pmp/iop-pmp-nub"].energy_model_dram_configs
        self.asc.iface.writemem(self.shmem + 0x2000, dram_config)
        
        node = self.asc.u.adt["arm-io/pmp"]
        
        maps = []
        dva = 0xc0000000
        for i in range(3, len(node.reg)):
            addr, size = node.get_reg(i)
            if size == 0:
                maps.append(struct.pack("<QQ", 0, 0))
                continue

            self.asc.dart.iomap_at(0, dva, addr, size)
            self.log(f"map {addr:#x} -> {dva:#x} [{size:#x}]")
            maps.append(struct.pack("<QQ", dva, size))
            dva += align(size, 0x4000)

        chexdump(b"".join(maps))

        self.asc.iface.writemem(self.shmem + 0xe000, b"".join(maps))
        self.send(PMP_Configure(DVA=self.shmem_dva))

        while not self.init_complete:
            self.asc.work()
        return True

    @msg_handler(0x20, PMP_Configure_Ack)
    def Configure_Ack(self, msg):
        self.init_complete = True

        props = self.asc.iface.readmem(self.shmem, 0x2000)
        devinfo = self.asc.iface.readmem(self.shmem + 0x4000, 0x1000)
        status = self.asc.iface.readmem(self.shmem + 0xc000, 0x100)

        print("PMP Props:")
        chexdump(props)
        print("PMP Device Info:")
        chexdump(devinfo)
        print("PMP Status:")
        chexdump(status)

        self.send(PMP_Init1(UNK1=1, UNK2=3))
        while not self.init1_acked:
            self.asc.work()

        self.send(PMP_Init2(UNK1=1, UNK2=0))
        while not self.init2_acked:
            self.asc.work()

        self.send(PMP_Unk(UNK1=0x3bc, UNK2=2))
        while not self.unk_acked:
            self.asc.work()

        return True

    @msg_handler(0x201, PMP_Init1_Ack)
    def Init1_Ack(self, msg):
        self.init1_acked = True
        return True

    @msg_handler(0x203, PMP_Init2_Ack)
    def Init2_Ack(self, msg):
        self.init2_acked = True
        return True

    @msg_handler(0x110, PMP_Unk_Ack)
    def Unk_Ack(self, msg):
        self.unk_acked = True
        return True


class PMPClient(StandardASC):
    pass

    ENDPOINTS = {
        0x20: PMPEndpoint,
    }



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import atexit, sys

from m1n1.setup import *
from m1n1.constructutils import Ver
from m1n1.utils import *

Ver.set_version(u)

from m1n1.agx import AGX
from m1n1.agx.render import *

from m1n1 import asm

p.pmgr_adt_clocks_enable("/arm-io/gfx-asc")
p.pmgr_adt_clocks_enable("/arm-io/sgx")

agx = AGX(u)
mon = RegMonitor(u, ascii=True, bufsize=0x8000000)
agx.mon = mon

sgx = agx.sgx_dev

try:
    agx.start()
    agx.uat.dump(0)

    print("==========================================")
    print("## After init")
    print("==========================================")
    mon.poll()
    agx.poll_objects()

    ctx = GPUContext(agx)
    ctx.bind(63)

    f = GPUFrame(ctx, sys.argv[1], track=False)

    r = GPURenderer(ctx, 16, bm_slot=0, queue=1)
    print("==========================================")
    print("## Submitting")
    print("==========================================")

    w = r.submit(f.cmdbuf)

    print("==========================================")
    print("## Submitted")
    print("==========================================")

    print("==========================================")
    print("## Run")
    print("==========================================")

    r.run()

    while not r.ev_3d.fired:
        agx.asc.work()
        agx.poll_channels()

    agx.poll_objects()
    mon.poll()
    r.wait()

    time.sleep(3)

finally:
    #agx.poll_objects()
    p.reboot()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *

blacklist = []

print("Dumping address space...")
of = None
if len(sys.argv) > 1:
	of = open(sys.argv[1],"w")
	print("Also dumping to file %s")

for i in range(0x0000, 0x10000):
	if i in blacklist:
		v = "%08x: SKIPPED"%(i<<16)
	else:
		a = (i<<16) + 0x1000
		d = p.read32(a)
		v = "%08x: %08x"%(a, d)
	print(v)
	if of:
		of.write(v+"\n")

if of:
	of.close()



# SPDX-License-Identifier: MIT

from m1n1.hv import TraceMode
from m1n1.hw.dwc3 import XhciRegs, Dwc3CoreRegs, PipehandlerRegs
from m1n1.hw.atc import Usb2PhyRegs, AtcPhyRegs
from m1n1.trace import ADTDevTracer
from m1n1.utils import *


class PhyTracer(ADTDevTracer):
    REGMAPS = [
        Usb2PhyRegs,
        None,
        (AtcPhyRegs, 0x20000),
        (AtcPhyRegs, 0x0),
        (AtcPhyRegs, 0x2000),
        (AtcPhyRegs, 0x2200),
        (AtcPhyRegs, 0x2800),
        (AtcPhyRegs, 0x2A00),
        (AtcPhyRegs, 0x7000),
        (AtcPhyRegs, 0xA00),
        (AtcPhyRegs, 0x800),
        (AtcPhyRegs, 0xD000),
        (AtcPhyRegs, 0x14000),
        (AtcPhyRegs, 0xC000),
        (AtcPhyRegs, 0x13000),
        (AtcPhyRegs, 0xB000),
        (AtcPhyRegs, 0x12000),
        (AtcPhyRegs, 0x9000),
        (AtcPhyRegs, 0x10000),
        (AtcPhyRegs, 0x1000),
        (AtcPhyRegs, 0x50000),
        (AtcPhyRegs, 0x50200),
        (AtcPhyRegs, 0x54000),
        None,
        None,
        None,
        (AtcPhyRegs, 0xA000),
        (AtcPhyRegs, 0x11000),
    ]


class Dwc3VerboseTracer(ADTDevTracer):
    REGMAPS = [XhciRegs, None, Dwc3CoreRegs, PipehandlerRegs]
    NAMES = ["xhci", None, "dwc-core", "pipehandler"]


class Dwc3Tracer(ADTDevTracer):
    REGMAPS = [None, None, Dwc3CoreRegs, PipehandlerRegs]
    NAMES = [None, None, "dwc-core", "pipehandler"]


PhyTracer = PhyTracer._reloadcls()
Dwc3Tracer = Dwc3Tracer._reloadcls()
Dwc3VerboseTracer = Dwc3VerboseTracer._reloadcls()

phy_tracer = PhyTracer(hv, "/arm-io/atc-phy1", verbose=2)
dwc3_tracer = Dwc3Tracer(hv, "/arm-io/usb-drd1", verbose=2)

phy_tracer.start()
dwc3_tracer.start()



# SPDX-License-Identifier: MIT
from m1n1.utils import align_up

def collect_aic_irqs_in_use(adt):
	used = set()
	aic_phandle = getattr(adt["/arm-io/aic"], "AAPL,phandle")
	for node in adt.walk_tree():
		if not hasattr(node, "interrupt_parent") or \
				node.interrupt_parent != aic_phandle:
			continue
		for no in node.interrupts:
			used.add(no)
	return used

def usable_aic_irq_range(adt):
	# These are too optimistic but since we allocate
	# from the bottom of the range it doesn't matter much.
	return {
		"aic,1": range(0, 0x400),
		"aic,2": range(0, 0x1000),
	}.get(adt["/arm-io/aic"].compatible[0])

def alloc_aic_irq(adt):
	used = collect_aic_irqs_in_use(adt)
	for no in usable_aic_irq_range(adt):
		if no not in used:
			return no
	return None

def usable_mmio_range(adt):
	arm_io_range = adt["arm-io"].ranges[0]
	return range(arm_io_range.parent_addr, arm_io_range.parent_addr + arm_io_range.size)

def alloc_mmio_base(adt, size, alignment=0x4000):
	span = usable_mmio_range(adt)
	la = adt.build_addr_lookup()
	for zone, devs in la.populate(span):
		if len(devs) != 0:
			continue
		base = align_up(zone.start, alignment)
		if zone.stop > base + size:
			return base
	return None



from m1n1.utils import RegMap, Register8
from enum import IntEnum
from .cs42l84 import *
from .ssm3515 import *

class E_PWR_MODE(IntEnum):
    ACTIVE   = 0
    MUTE     = 1
    SHUTDOWN = 2

class R_PWR_CTL(Register8):
    ISNS_PD = 3
    VSNS_PD = 2
    MODE    = 1, 0, E_PWR_MODE

class R_PB_CFG0(Register8):
    PDM_MAP    = 7
    PB_PDM_SRC = 6
    PB_SRC     = 5
    AMP_LEVEL  = 4, 0

class R_PB_CFG2(Register8):
    DVC_PCM = 7, 0

class R_PB_CFG3(Register8):
    DVC_PDM = 7, 0

class E_RX_SCFG(IntEnum):
    I2C_OFFSET = 0b00
    LEFT       = 0b01
    RIGHT      = 0b10
    DOWNMIX    = 0b11

class E_RX_WLEN(IntEnum):
    W_16BIT = 0b00
    W_20BIT = 0b01
    W_24BIT = 0b10
    W_32BIT = 0b11

class E_RX_SLEN(IntEnum):
    W_16BIT = 0b00
    W_24BIT = 0b01
    W_32BIT = 0b10

class R_TDM_CFG2(Register8):
    RX_SCFG = 5, 4, E_RX_SCFG
    RX_WLEN = 3, 2, E_RX_WLEN
    RX_SLEN = 1, 0, E_RX_SLEN

class R_TDM_CFG3(Register8):
    RX_SLOT_R = 7, 4
    RX_SLOT_L = 3, 0

class TAS5770Regs(RegMap):
    PWR_CTL  = 0x002, R_PWR_CTL
    PB_CFG0  = 0x003, R_PB_CFG0
    PB_CFG2  = 0x005, R_PB_CFG2
    PB_CFG3  = 0x006, R_PB_CFG3
    TDM_CFG2 = 0x00c, R_TDM_CFG2
    TDM_CFG3 = 0x00d, R_TDM_CFG3

class R_MODE_CTRL(Register8):
    BOP_SRC = 7
    ISNS_PD = 3
    VSNS_PD = 2
    MODE    = 1, 0, E_PWR_MODE

class R_CHNL_0(Register8):
    CDS_MODE  = 7, 6
    AMP_LEVEL = 5, 1

class R_DVC(Register8):
    DVC_LVL = 7, 0

class R_INT_MASK0(Register8):
    BOPM  = 7
    BOPIH = 6
    LIMMA = 5
    PBIP  = 4
    LIMA  = 3
    TDMCE = 2
    OC    = 1
    OT    = 0

class R_INT_CLK_CFG(Register8):
    CLK_ERR_PWR_EN = 7
    DIS_CLK_HAT    = 6
    CLK_HALT_TIMER = 5, 3
    IRQZ_CLR       = 2
    IRQZ_PIN_CFG   = 1, 0

class SN012776Regs(RegMap):
    MODE_CTRL = 0x002, R_MODE_CTRL
    CHNL_0    = 0x003, R_CHNL_0
    DVC       = 0x01a, R_DVC

    INT_MASK0 = 0x03b, R_INT_MASK0
    INT_MASK1 = 0x03c, Register8
    INT_MASK2 = 0x040, Register8
    INT_MASK3 = 0x041, Register8
    INT_MASK4 = 0x03d, Register8

    INT_LTCH0   = 0x049, R_INT_MASK0
    INT_LTCH1   = 0x04a, Register8
    INT_LTCH1_0 = 0x04b, Register8
    INT_LTCH2   = 0x04f, Register8
    INT_LTCH3   = 0x050, Register8
    INT_LTCH4   = 0x051, Register8

    INT_CLK_CFG = 0x05c, R_INT_CLK_CFG



# SPDX-License-Identifier: MIT
from ..utils import *
import time

class R_MBOX_CTRL(Register32):
    FIFOCNT = 23, 20
    OVERFLOW = 18
    EMPTY = 17
    FULL  = 16
    RPTR = 15, 12
    WPTR = 11, 8
    ENABLE = 0

class R_CPU_CONTROL(Register32):
    RUN    = 4

class R_CPU_STATUS(Register32):
    IDLE            = 5
    FIQ_NOT_PEND    = 3 # guess
    IRQ_NOT_PEND    = 2 # guess
    STOPPED         = 1
    RUNNING         = 0

class R_INBOX1(Register64):
    EP      = 7, 0

class R_OUTBOX1(Register64):
    OUTCNT  = 56, 52
    INCNT   = 51, 48
    OUTPTR  = 47, 44
    INPTR   = 43, 40
    EP      = 7, 0

class ASCRegs(RegMap):
    CPU_CONTROL = 0x0044, R_CPU_CONTROL
    CPU_STATUS  = 0x0048, R_CPU_STATUS

    INBOX_CTRL  = 0x8110, R_MBOX_CTRL
    OUTBOX_CTRL = 0x8114, R_MBOX_CTRL
    INBOX0      = 0x8800, Register64
    INBOX1      = 0x8808, R_INBOX1
    OUTBOX0     = 0x8830, Register64
    OUTBOX1     = 0x8838, R_OUTBOX1

class ASC:
    def __init__(self, u, asc_base):
        self.u = u
        self.p = u.proxy
        self.iface = u.iface
        self.asc = ASCRegs(u, asc_base)
        self.verbose = 0
        self.epmap = {}

    def recv(self):
        if self.asc.OUTBOX_CTRL.reg.EMPTY:
            return None, None

        msg0 = self.asc.OUTBOX0.val
        msg1 = R_INBOX1(self.asc.OUTBOX1.val)
        if self.verbose >= 3:
            print(f"< {msg1.EP:02x}:{msg0:#x}")
        return msg0, msg1

    def send(self, msg0, msg1):
        self.asc.INBOX0.val = msg0
        self.asc.INBOX1.val = msg1

        if self.verbose >= 3:
            if isinstance(msg0, Register):
                print(f"> {msg1.EP:02x}:{msg0}")
            else:
                print(f"> {msg1.EP:02x}:{msg0:#x}")

        while self.asc.INBOX_CTRL.reg.FULL:
            pass

    def is_running(self):
        return not self.asc.CPU_STATUS.reg.STOPPED

    def boot(self):
        self.asc.CPU_CONTROL.set(RUN=1)

    def shutdown(self):
        self.asc.CPU_CONTROL.set(RUN=0)

    def add_ep(self, idx, ep):
        self.epmap[idx] = ep
        setattr(self, ep.SHORT, ep)

    def has_messages(self):
        return not self.asc.OUTBOX_CTRL.reg.EMPTY

    def work_pending(self):
        while self.has_messages():
            self.work()

    def work(self):
        if self.asc.OUTBOX_CTRL.reg.EMPTY:
            return True

        msg0, msg1 = self.recv()

        handled = False

        ep = self.epmap.get(msg1.EP, None)
        if ep:
            handled = ep.handle_msg(msg0, msg1)

        if not handled:
            print(f"unknown message: {msg0:#16x} / {msg1}")

        return handled

    def work_forever(self):
        while self.work():
            pass

    def work_for(self, timeout):
        deadline = time.time() + timeout
        while time.time() < deadline:
            self.work()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import inspect, textwrap, json, re, sys, os

from construct import *
from construct.core import evaluate
from construct.lib import HexDisplayedInteger
from .utils import *

g_struct_trace = set()
g_struct_addrmap = {}
g_depth = 0

def ZPadding(size):
    return Const(bytes(size), Bytes(size))

def recursive_reload(obj, token=None):
    global g_depth

    if token is None:
        g_depth = 0
        token = object()

    cur_token = getattr(obj, "_token", None)
    if cur_token is token:
        return

    g_depth += 1
    #print("  " * g_depth + f"> {obj}", id(obj), id(token))
    if isinstance(obj, Construct) and hasattr(obj, 'subcon'):
        # Single subcon types
        if inspect.isclass(obj.subcon):
            #print("> isclass")
            if hasattr(obj.subcon, "_reloadcls"):
                #print("> Recursive (subcon)")
                obj.subcon = obj.subcon._reloadcls(token=token)
        else:
            if isinstance(obj.subcon, Construct):
                recursive_reload(obj.subcon, token)
    if isinstance(obj, Construct) and hasattr(obj, 'subcons'):
        # Construct types that have lists
        new_subcons = []
        for i, item in enumerate(obj.subcons):
            if inspect.isclass(item):
                if hasattr(item, "_reloadcls"):
                    #print("> Recursive (subcons)")
                    item = item._reloadcls()
            else:
                if isinstance(item, Construct):
                    recursive_reload(item, token)
            new_subcons.append(item)
            obj.subcons = new_subcons

    if isinstance(obj, Construct) and hasattr(obj, 'cases'):
        # Construct types that have lists
        for i, item in list(obj.cases.items()):
            if inspect.isclass(item):
                if hasattr(item, "_reloadcls"):
                    #print("> Recursive (cases)")
                    obj.cases[i] = item._reloadcls(token=token)
            else:
                if isinstance(item, Construct):
                    recursive_reload(item, token)

    for field in dir(obj):
        value = getattr(obj, field)
        if inspect.isclass(value):
            if hasattr(value, "_reloadcls"):
                #print("> Recursive (value)")
                setattr(obj, field, value._reloadcls(token=token))
        else:
            if isinstance(value, Construct):
                recursive_reload(value, token)

    obj._token = token

    g_depth -= 1

def str_value(value, repr=False):
    if isinstance(value, bytes) and value == bytes(len(value)):
        return f"bytes({len(value):#x})"
    if isinstance(value, bytes) and repr:
        return f"bytes.fromhex('{value.hex()}')"
    if isinstance(value, DecDisplayedInteger):
        return str(value)
    if isinstance(value, int):
        if value in g_struct_addrmap:
            desc = g_struct_addrmap[value]
            return f"{value:#x} ({desc})"
        else:
            return f"{value:#x}"
    if isinstance(value, ListContainer):
        om = ""
        while len(value) > 1 and not value[-1]:
            value = value[:-1]
            om = " ..."
        if len(value) <= 16:
            return "[" + ", ".join(map(str_value, value)) + f"{om}]"
        else:
            sv = ["[\n"]
            for off in range(0, len(value), 16):
                sv.append("  " + ", ".join(map(str_value, value[off:off+16])) + ",\n")
            sv.append(f"{om}]\n")
            return "".join(sv)

    return str(value)

class DecDisplayedInteger(int):
    @staticmethod
    def new(intvalue):
        obj = DecDisplayedInteger(intvalue)
        return obj

class Dec(Adapter):
    def _decode(self, obj, context, path):
        try:
            if isinstance(obj, int):
                return DecDisplayedInteger.new(obj)
            return obj
        except Exception as e:
            print(e)
            raise

    def _encode(self, obj, context, path):
        return obj

    def _emitparse(self, code):
        return self.subcon._compileparse(code)

    def _emitseq(self, ksy, bitwise):
        return self.subcon._compileseq(ksy, bitwise)

    def _emitprimitivetype(self, ksy, bitwise):
        return self.subcon._compileprimitivetype(ksy, bitwise)

    def _emitfulltype(self, ksy, bitwise):
        return self.subcon._compilefulltype(ksy, bitwise)

class ConstructClassException(Exception):
    pass


# We need to inherit Construct as a metaclass so things like If and Select will work
class ReloadableConstructMeta(ReloadableMeta, Construct):

    def __new__(cls, name, bases, attrs):
        cls = super().__new__(cls, name, bases, attrs)
        cls.name = name
        if cls.SHORT_NAME is not None:
            cls.short_name = cls.SHORT_NAME
        else:
            cls.short_name = re.sub('[a-z]', '', cls.name)
            if len(cls.short_name) > 5:
                cls.short_name = cls.short_name[:3] + cls.short_name[-2:]

        try:
            cls.flagbuildnone = cls.subcon.flagbuildnone
        except AttributeError:
            cls.flagbuildnone = False

        cls.docs = None

        cls._off = {}
        if "subcon" not in attrs:
            return cls

        subcon = attrs["subcon"]
        if isinstance(subcon, Struct):
            off = 0
            for subcon in subcon.subcons:
                try:
                    sizeof = subcon.sizeof()
                except:
                    sizeof = None
                if isinstance(subcon, Ver):
                    if not subcon._active():
                        cls._off[subcon.name] = -1, 0
                        continue
                    subcon = subcon.subcon
                if isinstance(subcon, Renamed):
                    name = subcon.name
                    subcon = subcon.subcon
                    cls._off[name] = off, sizeof
                if sizeof is None:
                    break
                off += sizeof
        return cls

class ConstructClassBase(Reloadable, metaclass=ReloadableConstructMeta):
    """ Offers two benefits over regular construct

        1. It's reloadable, and can recursively reload other referenced ConstructClasses
        2. It's a class, so you can define methods

        Currently only supports parsing, but could be extended to support building

        Example:
            Instead of:
            MyStruct = Struct(
                "field1" / Int32ul
            )

            class MyClass(ConstructClass):
                subcon = Struct(
                    "field1" / Int32ul
                )

    """
    SHORT_NAME = None

    parsed = None

    def __init__(self):
        self._pointers = set()
        self._addr = None
        self._meta = {}

    def regmap(self):
        return ConstructRegMap(type(self), self._stream.to_accessor(), self._addr)

    @classmethod
    def sizeof(cls, **contextkw):
        context = Container(**contextkw)
        context._parsing = False
        context._building = False
        context._sizing = True
        context._params = context
        return cls._sizeof(context, "(sizeof)")

    def Apply(self, dict=None, **kwargs):
        if dict is None:
            dict = kwargs

        for key in dict:
            if not key.startswith('_'):
                setattr(self, key, dict[key])
                self._keys += [key]

    def set_addr(self, addr=None, stream=None):
        #print("set_addr", type(self), addr)
        if addr is not None:
            self._addr = addr
        self._set_meta(self, stream)

    @classmethod
    def _build(cls, obj, stream, context, path):
        cls._build_prepare(obj)

        addr = stream.tell()
        try:
            new_obj = cls.subcon._build(obj, stream, context, f"{path} -> {cls.name}")
        except ConstructClassException:
            raise
        except ConstructError:
            raise
        except Exception as e:
            raise ConstructClassException(f"at {path} -> {cls.name}") from e

        # if obj is a raw value or Container, instance a proper object for it
        if not isinstance(obj, ConstructClassBase):
            obj = cls.__new__(cls)

        # update the object with anything that build updated (such as defaults)
        obj._apply(new_obj)

        obj._addr = addr
        cls._set_meta(obj, stream)
        return obj

    @classmethod
    def _sizeof(cls, context, path):
        return cls.subcon._sizeof(context, f"{path} -> {cls.name}")

    @classmethod
    def _reloadcls(cls, force=False, token=None):
        #print(f"_reloadcls({cls})", id(cls))
        newcls = Reloadable._reloadcls.__func__(cls, force)
        if hasattr(newcls, "subcon"):
            recursive_reload(newcls.subcon, token)
        return newcls

    def _apply(self, obj):
        raise NotImplementedError()

    @classmethod
    def _set_meta(cls, self, stream=None):
        if stream is not None:
            self._pointers = set()
            self._meta = {}
            self._stream = stream

        if isinstance(cls.subcon, Struct):
            subaddr = int(self._addr)
            for subcon in cls.subcon.subcons:
                try:
                    sizeof = subcon.sizeof()
                except:
                    break
                if isinstance(subcon, Ver):
                    subcon = subcon.subcon
                if isinstance(subcon, Renamed):
                    name = subcon.name
                    #print(name, subcon)
                    subcon = subcon.subcon
                    if stream is not None and getattr(stream, "meta_fn", None):
                        meta = stream.meta_fn(subaddr, sizeof)
                        if meta is not None:
                            self._meta[name] = meta
                    if isinstance(subcon, Pointer):
                        self._pointers.add(name)
                        continue
                    try:
                        #print(name, subcon)
                        val = self[name]
                    except:
                        pass
                    else:
                        if isinstance(val, ConstructClassBase):
                            val.set_addr(subaddr)
                        if isinstance(val, list):
                            subaddr2 = subaddr
                            for i in val:
                                if isinstance(i, ConstructClassBase):
                                    i.set_addr(subaddr2)
                                    subaddr2 += i.sizeof()

                subaddr += sizeof

    @classmethod
    def _parse(cls, stream, context, path):
        #print(f"parse {cls} @ {stream.tell():#x} {path}")
        addr = stream.tell()
        obj = cls.subcon._parse(stream, context, path)
        size = stream.tell() - addr

        # Don't instance Selects
        if isinstance(cls.subcon, Select):
            return obj

        # Skip calling the __init__ constructor, so that it can be used for building
        # Use parsed instead, if you need a post-parsing constructor
        self = cls.__new__(cls)
        self._addr = addr
        self._path = path
        self._meta = {}
        cls._set_meta(self, stream)

        self._apply(obj)

        if self._addr > 0x10000:
            desc = f"{cls.name} (end: {self._addr + size:#x})"
            if getattr(stream, "meta_fn", None):
                meta = stream.meta_fn(self._addr, None)
                if meta is not None:
                    desc += " " + meta
            g_struct_trace.add((self._addr, desc))
            g_struct_addrmap[self._addr] = f"{cls.name}"
        return self

    @classmethod
    def _build_prepare(cls, obj):
        pass

    def build_stream(self, obj=None, stream=None, **contextkw):
        assert stream != None
        if obj is None:
            obj = self

        return Construct.build_stream(self, obj, stream, **contextkw)

    def build(self, obj=None, **contextkw):
        if obj is None:
            obj = self

        return Construct.build(self, obj, **contextkw)

class ROPointer(Pointer):
    def _build(self, obj, stream, context, path):
        return obj

    def _parse(self, stream, context, path):
        recurse = getattr(stream, "recurse", False)
        if not recurse:
            return None

        return Pointer._parse(self, stream, context, path)

class ConstructClass(ConstructClassBase, Container):
    """ Offers two benefits over regular construct

        1. It's reloadable, and can recursively reload other referenced ConstructClasses
        2. It's a class, so you can define methods

        Currently only supports parsing, but could be extended to support building

        Example:
            Instead of:
            MyStruct = Struct(
                "field1" / Int32ul
            )

            class MyClass(ConstructClass):
                subcon = Struct(
                    "field1" / Int32ul
                )
    """

    def diff(self, other, show_all=False):
        return self.__str__(other=other, show_all=show_all)

    def __eq__(self, other):
        return all(self[k] == other[k] for k in self
                   if (not k.startswith("_"))
                   and (k not in self._pointers)
                   and not callable(self[k]))

    def __str__(self, ignore=[], other=None, show_all=False) -> str:

        str = self.__class__.__name__
        if self._addr is not None:
            str += f" @ 0x{self._addr:x}:"

        str += "\n"

        keys = list(self)
        keys.sort(key = lambda x: self._off.get(x, (-1, 0))[0])

        for key in keys:
            if key in self._off:
                offv, sizeof = self._off[key]
                if offv == -1:
                    # print(key, offv, sizeof)
                    continue
            if key in ignore or key.startswith('_'):
                continue
            value = getattr(self, key)
            need_diff = False
            if other is not None:
                if key in self._pointers or callable(value):
                    continue
                other_value = getattr(other, key)
                if not show_all and other_value == value:
                    continue
                offv, sizeof = self._off[key]
                if sizeof == 0:
                    continue
                def _valdiff(value, other_value):
                    if hasattr(value, "diff"):
                        return value.diff(other_value)
                    elif isinstance(value, bytes) and isinstance(other_value, bytes):
                        pad = bytes()
                        if len(value) & 3:
                            pad = bytes(4 - (len(value) & 3))
                        return chexdiff32(other_value+pad, value+pad, offset=offv, offset2=0)
                    else:
                        val_repr = str_value(value)
                        if other_value != value:
                            other_repr = str_value(other_value)
                            return f"\x1b[33;1;4m{val_repr}\x1b[m ← \x1b[34m{other_repr}\x1b[m"
                        return val_repr

                if isinstance(value, list):
                    val_repr = "{\n"
                    for i, (a, b) in enumerate(zip(value, other_value)):
                        if a == b:
                            continue
                        val_repr += f"[{i}] = " + textwrap.indent(_valdiff(a, b), "    ") + "\n"
                        offv += sizeof // len(value)
                    val_repr += "}\n"
                else:
                    val_repr = _valdiff(value, other_value)

            else:
                val_repr = str_value(value)
            off = ""
            meta = ""
            if key in self._off:
                offv, sizeof = self._off[key]
                if sizeof is not None:
                    sizeofs = f"{sizeof:3x}"
                else:
                    sizeofs = "  *"
                off = f"\x1b[32m[{offv:3x}.{sizeofs}]\x1b[m "
            if key in self._meta:
                meta = f" \x1b[34m{self._meta[key]}\x1b[m"
            if '\n' in val_repr:
                val_repr = textwrap.indent(val_repr, f'\x1b[90m{self.short_name:>5s}.\x1b[m')
                if not val_repr.endswith('\n'):
                    val_repr += '\n'
                str += f"\x1b[90m{self.short_name:>5s}.{off}\x1b[95m{key}\x1b[m ={meta}\n{val_repr}"
            else:
                str += f"\x1b[90m{self.short_name:>5s}.{off}\x1b[95m{key}\x1b[m = {val_repr}{meta}\n"

        return str

    def _dump(self):
        print(f"# {self.__class__.__name__}")
        if self._addr is not None:
            print(f"#  Address: 0x{self._addr:x}")

        keys = list(self)
        keys.sort(key = lambda x: self._off.get(x, (-1, 0))[0])
        for key in keys:
            if key.startswith('_'):
                continue
            value = getattr(self, key)
            val_repr = str_value(value, repr=True)
            print(f"self.{key} = {val_repr}")


    @classmethod
    def _build_prepare(cls, obj):
        if isinstance(cls.subcon, Struct):
            for subcon in cls.subcon.subcons:
                if isinstance(subcon, Ver):
                    subcon = subcon.subcon
                if not isinstance(subcon, Renamed):
                    continue
                name = subcon.name
                subcon = subcon.subcon
                if isinstance(subcon, Lazy):
                    subcon = subcon.subcon
                if not isinstance(subcon, Pointer):
                    continue
                addr_field = subcon.offset.__getfield__()
                # Ugh.
                parent = subcon.offset._Path__parent(obj)
                if not hasattr(obj, name) and hasattr(parent, addr_field):
                    # No need for building
                    setattr(obj, name, None)
                elif hasattr(obj, name):
                    subobj = getattr(obj, name)
                    try:
                        addr = subobj._addr
                    except (AttributeError, KeyError):
                        addr = None
                    if addr is not None:
                        setattr(parent, addr_field, addr)

    @classmethod
    def _parse(cls, stream, context, path):
        self = ConstructClassBase._parse.__func__(cls, stream, context, path)

        for key in self:
            if key.startswith('_'):
                continue
            try:
                val = int(self[key])
            except:
                continue
            if (0x1000000000 <= val <= 0x1f00000000 or
                0xf8000000000 <= val <= 0xff000000000 or
                0xffffff8000000000 <= val <= 0xfffffff000000000):
                g_struct_trace.add((val, f"{cls.name}.{key}"))
        return self

    def _apply_classful(self, obj):
        obj2 = dict(obj)
        if isinstance(self.__class__.subcon, Struct):
            for subcon in self.__class__.subcon.subcons:
                name = subcon.name
                if name is None:
                    continue
                subcon = subcon.subcon
                if isinstance(subcon, Lazy):
                    continue
                if isinstance(subcon, Pointer):
                    subcon = subcon.subcon
                if isinstance(subcon, Ver):
                    subcon = subcon.subcon
                if isinstance(subcon, Array):
                    subcon = subcon.subcon

                if name not in obj2:
                    continue

                val = obj2[name]
                if not isinstance(subcon, type) or not issubclass(subcon, ConstructClassBase):
                    continue

                def _map(v):
                    if not isinstance(v, subcon):
                        sc = subcon()
                        sc._apply(v)
                        return sc
                    return v

                if isinstance(val, list):
                    obj2[name] = list(map(_map, val))
                else:
                    obj2[name] = _map(val)

        self._apply(obj2)

    def _apply(self, obj):
        self.update(obj)

    def items(self):
        for k in list(self):
            if k.startswith("_"):
                continue
            v = self[k]
            if getattr(v, "HAS_VALUE", None):
                yield k, v.value
            else:
                yield k, v

    def addrof(self, name):
        return self._addr + self._off[name][0]

    @classmethod
    def offsetof(cls, name):
        return cls._off[name][0]

    def clone(self):
        obj = type(self)()
        obj.update(self)
        return obj

    @classmethod
    def from_json(cls, fd):
        d = json.load(fd)
        obj = cls()
        obj._apply_classful(d)
        return obj

    @classmethod
    def is_versioned(cls):
        for subcon in cls.subcon.subcons:
            if isinstance(subcon, Ver):
                return True
            while True:
                try:
                    subcon = subcon.subcon
                    if isinstance(subcon, type) and issubclass(subcon, ConstructClass) and subcon.is_versioned():
                        return True
                except:
                    break
        return False

    @classmethod
    def to_rust(cls):
        assert isinstance(cls.subcon, Struct)
        s = []
        if cls.is_versioned():
            s.append("#[versions(AGX)]"),

        s += [
            "#[derive(Debug, Clone, Copy)]",
            "#[repr(C, packed(4))]",
            f"struct {cls.__name__} {{",
        ]
        pad = 0
        has_ver = False
        for subcon in cls.subcon.subcons:
            if isinstance(subcon, Ver):
                if not has_ver:
                    s.append("")
                s.append(f"    #[ver({subcon.cond})]")
                subcon = subcon.subcon
                has_ver = True
            else:
                has_ver = False

            name = subcon.name
            if name is None:
                name = f"__pad{pad}"
                pad += 1

            array_len = []
            skip = False
            while subcon:
                if isinstance(subcon, Lazy):
                    skip = True
                    break
                elif isinstance(subcon, Pointer):
                    skip = True
                    break
                elif isinstance(subcon, Array):
                    array_len.append(subcon.count)
                elif isinstance(subcon, (HexDump, Default, Renamed, Dec, Hex, Const)):
                    pass
                else:
                    break
                subcon = subcon.subcon

            if isinstance(subcon, Bytes):
                array_len.append(subcon.length)
                subcon = Int8ul

            if skip:
                #s.append(f"    // {name}: {subcon}")
                continue

            TYPE_MAP = {
                Int64ul: "u64",
                Int32ul: "u32",
                Int16ul: "u16",
                Int8ul: "u8",
                Int64sl: "i64",
                Int32sl: "i32",
                Int16sl: "i16",
                Int8sl: "i8",
                Float32l: "f32",
                Float64l: "f64",
            }

            t = TYPE_MAP.get(subcon, repr(subcon))

            if isinstance(subcon, type) and issubclass(subcon, ConstructClass):
                t = subcon.__name__
                if subcon.is_versioned():
                    t += "::ver"

            for n in array_len[::-1]:
                t = f"Array<{n:#x}, {t}>"

            s.append(f"    pub(crate) {name}: {t},")

            if has_ver:
                s.append("")

        s += ["}"]
        return "\n".join(s)

class ConstructValueClass(ConstructClassBase):
    """ Same as Construct, but for subcons that are single values, rather than containers

        the value is stored as .value
    """
    HAS_VALUE = True

    def __eq__(self, other):
        return self.value == other.value

    def __str__(self) -> str:
        str = f"{self.__class__.__name__} @ 0x{self._addr:x}:"
        str += f"\t{str_value(self.value)}"
        return str

    def __getitem__(self, i):
        if i == "value":
            return self.value
        raise Exception(f"Invalid index {i}")

    @classmethod
    def _build(cls, obj, stream, context, path):
        return super()._build(obj.value, stream, context, path)

    def _apply(self, obj):
        self.value = obj
    _apply_classful = _apply

class ConstructRegMap(BaseRegMap):
    TYPE_MAP = {
        Int8ul: Register8,
        Int16ul: Register16,
        Int32ul: Register32,
        Int64ul: Register64,
    }

    def __init__(self, cls, backend, base):
        self._addrmap = {}
        self._rngmap = SetRangeMap()
        self._namemap = {}
        assert isinstance(cls.subcon, Struct)
        for subcon in cls.subcon.subcons:
            if isinstance(subcon, Ver):
                subcon = subcon.subcon
            if not isinstance(subcon, Renamed):
                continue
            name = subcon.name
            subcon = subcon.subcon
            if subcon not in self.TYPE_MAP:
                continue
            rtype = self.TYPE_MAP[subcon]
            if name not in cls._off:
                continue
            addr, size = cls._off[name]
            self._addrmap[addr] = name, rtype
            self._namemap[name] = addr, rtype
        super().__init__(backend, base)

    def __getattr__(self, k):
        if k.startswith("_"):
            return self.__dict__[k]
        return self._accessor[k]

    def __setattr__(self, k, v):
        if k.startswith("_"):
            self.__dict__[k] = v
            return
        self._accessor[k].val = v

class Ver(Subconstruct):
    # Ugly hack to make this survive across reloads...
    try:
        _version = sys.modules["m1n1.constructutils"].Ver._version
    except (KeyError, AttributeError):
        _version = {"V": os.environ.get("AGX_FWVER", "V12_3"),
                    "G": os.environ.get("AGX_GPU", "G13")}

    MATRIX = {
        "V": ["V12_1", "V12_3", "V12_4", "V13_0B4", "V13_0B5", "V13_0B6", "V13_2", "V13_3", "V13_5B4", "V13_5"],
        "G": ["G13", "G14", "G14X"],
    }

    def __init__(self, version, subcon):
        self.cond = version
        self.vcheck = self.parse_ver(version)
        self._name = subcon.name
        self.subcon = subcon
        self.flagbuildnone = True
        self.docs = ""
        self.parsed = None

    @property
    def name(self):
        if self._active():
            return self._name
        else:
            return None

    @staticmethod
    def _split_ver(s):
        if not s:
            return None
        parts = re.split(r"[-,. ]", s)
        parts2 = []
        for i in parts:
            try:
                parts2.append(int(i))
            except ValueError:
                parts2.append(i)
        if len(parts2) > 3 and parts2[-2] == "beta":
            parts2[-3] -= 1
            parts2[-2] = 99
        return tuple(parts2)

    @classmethod
    def parse_ver(cls, version):
        expr = version.replace("&&", " and ").replace("||", " or ")

        base_loc = {j: i for row in cls.MATRIX.values() for i, j in enumerate(row)}

        def check_ver(ver):
            loc = dict(base_loc)
            for k, v in ver.items():
                loc[k] = cls.MATRIX[k].index(v)
            return eval(expr, None, loc)

        return check_ver

    @classmethod
    def check(cls, version):
        return cls.parse_ver(version)(cls._version)

    def _active(self):
        return self.vcheck(self._version)

    def _parse(self, stream, context, path):
        if not self._active():
            return None
        obj = self.subcon._parse(stream, context, path)
        return obj

    def _build(self, obj, stream, context, path):
        if not self._active():
            return None
        return self.subcon._build(obj, stream, context, path)

    def _sizeof(self, context, path):
        if not self._active():
            return 0
        return self.subcon._sizeof(context, path)

    @classmethod
    def set_version_key(cls, key, version):
        cls._version[key] = version

    @classmethod
    def set_version(cls, u):
        cls.set_version_key("V", u.version)
        gpu = u.adt["/arm-io"].soc_generation.replace("H", "G")
        if gpu == "G14" and u.adt["/chosen"].chip_id < 0x8000:
            gpu = "G14X"
        cls.set_version_key("G", gpu)

def show_struct_trace(log=print):
    for addr, desc in sorted(list(g_struct_trace)):
        log(f"{addr:>#18x}: {desc}")

__all__ = ["ConstructClass", "ConstructValueClass", "Dec", "ROPointer", "show_struct_trace", "ZPadding", "Ver"]



# SPDX-License-Identifier: MIT
from ..utils import *
from enum import IntEnum
from .dart import DART
import struct
from enum import IntEnum


class AES_OPCODE(IntEnum):
    # 0 triggers an invalid command interrupt
    SET_KEY = 1
    SET_IV = 2
    # 0x03 seems to take three additional argument, function unknown
    # 0x04 seems to take one additional argument, function unknown
    CRYPT = 5
    GET_IV = 6
    # 0x07 takes one additional argument, function unknown
    BARRIER = 8  # can be used to trigger an IRQ but possibly also does more
    # > 8 trigger an invalid command interrupt


class AES_SET_KEY_LEN(IntEnum):
    AES128 = 0
    AES192 = 1
    AES256 = 2


class AES_SET_KEY_BLOCK_MODE(IntEnum):
    ECB = 0
    CBC = 1
    CTR = 2


class AESCommandBase(Register32):
    OPCODE = 31, 28, AES_OPCODE


class AESHwKey(IntEnum):
    SOFTWARE = 0
    UID = 1  # unique key for each chip
    GID0 = 2  # (probably) globally unique key within a chip family
    GID1 = 3  # globally unique key within a chip family
    # 4-7 are probably empty / reserved for future use


class AESSetKeyCommand(AESCommandBase):
    OPCODE = 31, 28, Constant(AES_OPCODE.SET_KEY)
    SLOT = 27, 27
    KEY_SELECT = 26, 24
    KEYLEN = 23, 22, AES_SET_KEY_LEN
    # setting bit 21 breaks the engine and sets two bits in the IRQ status
    ENCRYPT = 20, 20
    KEYGEN = 19, 18
    BLOCK_MODE = 17, 16, AES_SET_KEY_BLOCK_MODE
    # 15, 0 doesn't seem to have any effect


class AESCryptCommand(AESCommandBase):
    OPCODE = 31, 28, Constant(AES_OPCODE.CRYPT)
    KEY_SLOT = 27, 27
    IV_SLOT = 26, 25
    LEN = 24, 0


class AESBarrierCommand(AESCommandBase):
    OPCODE = 31, 28, Constant(AES_OPCODE.BARRIER)
    IRQ = 27, 27


class AESGetIVCommand(AESCommandBase):
    OPCODE = 31, 28, Constant(AES_OPCODE.GET_IV)


class AESSetIVCommand(AESCommandBase):
    OPCODE = 31, 28, Constant(AES_OPCODE.SET_IV)
    SLOT = 27, 26


class AESIrqReg(Register32):
    KEY1_EMPTY = 17, 17
    KEY1_INVALID = 13, 13
    KEY0_EMPTY = 11, 11
    KEY0_INVALID = 7, 7
    FLAG = 5, 5
    UNKNOWN_COMMAND = 2, 2
    FIFO_OVERFLOW = 1, 1


class AESControlReg(Register32):
    START = 0, 0
    STOP = 1, 1
    CLEAR_FIFO = 2, 2
    # TOOD: not convinced about RESET anymore, I remember this un-broke the engine once but I can't reproduce that anymore
    RESET = 3, 3


class AESFifoStatusReg(Register32):
    FIFO_WRITE_PTR = 31, 24
    FIFO_READ_PTR = 23, 16
    FIFO_LEVEL = 15, 8
    FIFO_FULL = 2, 2
    FIFO_EMPTY = 1, 1


class AESRegs(RegMap):
    R_CONTROL = 0x08, AESControlReg
    R_IRQ_STATUS = 0x18, AESIrqReg
    R_IRQ_ENABLE = 0x1C, AESIrqReg
    R_FIFO_STATUS = 0x24, AESFifoStatusReg
    R_CMD_FIFO = 0x200, Register32



# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from enum import IntEnum
from m1n1.utils import *
from construct import *

uint8_t = Int8ul
int16_t = Int16sl
uint16_t = Int16ul
int32_t = Int32sl
uint32_t = Int32ul
int64_t = Int64sl
uint64_t = Int64ul

uint = uint32_t
int_ = int32_t
ulong = uint64_t
long_ = int64_t

def Bool(c):
    return ExprAdapter(c, lambda d, ctx: bool(d & 1), lambda d, ctx: int(d))

def SizedArray(count, svar, subcon):
    return Padded(subcon.sizeof() * count, Array(lambda ctx: min(count, ctx.get(svar, ctx._.get(svar))), subcon))

def SizedBytes(count, svar):
    return Lazy(Padded(count, Bytes(lambda ctx: ctx.get(svar) or ctx._.get(svar))))

def UnkBytes(s):
    return Default(HexDump(Bytes(s)), b"\x00" * s)

bool_ = Bool(Int8ul)

class OSObject(Construct):
    TYPE = None

    def _parse(self, stream, context, path, recurse=False):
        tag = stream.read(1).decode("ascii")
        if not recurse and self.TYPE is not None and self.TYPE != tag:
            raise Exception("Object type mismatch")

        if tag == "d":
            count = Int32ul.parse_stream(stream)
            d = {}
            for i in range(count):
                k = self._parse(stream, context, path, True)
                v = self._parse(stream, context, path, True)
                d[k] = v
            return d
        elif tag == "n":
            return Int64ul.parse_stream(stream)
        elif tag == "s":
            length = Int32ul.parse_stream(stream)
            s = stream.read(length).decode("utf-8")
            assert stream.read(1) == b'\0'
            return s
        else:
            raise Exception(f"Unknown object tag {tag!r}")

    def _build(self, obj, stream, context, path):
        assert False

    def _sizeof(self, context, path):
        return None

class OSDictionary(OSObject):
    TYPE = 'd'

class OSSerialize(Construct):
    def _parse(self, stream, context, path, recurse=False):
        hdr = Int32ul.parse_stream(stream)
        if hdr != 0xd3:
            raise Exception("Bad header")

        obj, last = self.parse_obj(stream)
        assert last
        return obj

    def parse_obj(self, stream, level=0):
        # align to 32 bits
        pos = stream.tell()
        if pos & 3:
            stream.read(4 - (pos & 3))

        tag = Int32ul.parse_stream(stream)

        last = bool(tag & 0x80000000)
        otype = (tag >> 24) & 0x1f
        size = tag & 0xffffff

        #print(f"{'  '*level} @{stream.tell():#x} {otype} {last} {size}")

        if otype == 1:
            d = {}
            for i in range(size):
                k, l = self.parse_obj(stream, level + 1)
                assert not l
                v, l = self.parse_obj(stream, level + 1)
                assert l == (i == size - 1)
                d[k] = v
        elif otype == 2:
            d = []
            for i in range(size):
                v, l = self.parse_obj(stream, level + 1)
                assert l == (i == size - 1)
                d.append(v)
        elif otype == 4:
            d = Int64ul.parse_stream(stream)
        elif otype == 9:
            d = stream.read(size).decode("utf-8")
        elif otype == 10:
            d = stream.read(size)
        elif otype == 11:
            d = bool(size)
        else:
            raise Exception(f"Unknown tag {otype}")

        #print(f"{'  '*level}  => {d}")
        return d, last

    def build_obj(self, obj, stream, last=True, level=0):
        tag = 0
        if last:
            tag |= 0x80000000

        if isinstance(obj, dict):
            tag |= (1 << 24) | len(obj)
            Int32ul.build_stream(tag, stream)
            for i, (k, v) in enumerate(obj.items()):
                self.build_obj(k, stream, False, level + 1)
                self.build_obj(v, stream, i == len(obj) - 1, level + 1)
        elif isinstance(obj, list):
            tag |= (2 << 24) | len(obj)
            Int32ul.build_stream(tag, stream)
            for i, v in enumerate(obj):
                self.build_obj(v, stream, i == len(obj) - 1, level + 1)
        elif isinstance(obj, int):
            tag |= (4 << 24) | 64
            Int32ul.build_stream(tag, stream)
            Int64ul.build_stream(obj, stream)
        elif isinstance(obj, str):
            obj = obj.encode("utf-8")
            tag |= (9 << 24) | len(obj)
            Int32ul.build_stream(tag, stream)
            stream.write(obj)
        elif isinstance(obj, bytes):
            tag |= (10 << 24) | len(obj)
            Int32ul.build_stream(tag, stream)
            stream.write(obj)
        elif isinstance(obj, bool):
            tag |= (11 << 24) | int(obj)
            Int32ul.build_stream(tag, stream)
        else:
            raise Exception(f"Cannot encode {obj!r}")

        pos = stream.tell()
        if pos & 3:
            stream.write(bytes(4 - (pos & 3)))

    def _build(self, obj, stream, context, path):
        Int32ul.build_stream(0xd3, stream)
        self.build_obj(obj, stream)

    def _sizeof(self, context, path):
        return None

def string(size):
    return Padded(size, CString("utf8"))




#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1 import asm

FPCR_FZ = 1 << 24

ACTLR_DEFAULT = 0xc00
ACTLR_AFP = 1 << 5

AFPCR = (3,6,15,2,5)
AFPCR_DAZ = 1 << 0
AFPCR_FTZ = 1 << 1

code_buffer = p.malloc(0x1000)
data_buffer = p.malloc(0x1000)

code = asm.ARMAsm("""
    ldr s0, [x0, #0]
    ldr s1, [x0, #4]
    fmul s0, s1, s0
    str s0, [x0, #8]

    ldr s0, [x0, #12]
    ldr s1, [x0, #16]
    fmul s0, s1, s0
    str s0, [x0, #20]

    # to test EL0 access
    # mrs x0, s3_6_c15_c2_5
    ret
""", code_buffer)

iface.writemem(code_buffer, code.data)
p.dc_cvau(code_buffer, code.len)
p.ic_ivau(code_buffer, code.len)

def test_denormals():

    data = [
        0x00400000, # a denormal
        0x40000000, # 2
        0,
        0x00800000, # smallest non-denormal
        0x3f000000, # 0.5
        0,
    ]

    iface.writemem(data_buffer, struct.pack("<%dI" % len(data), *data))

    p.set_exc_guard(GUARD.SKIP)
    ret = p.el0_call(code_buffer, data_buffer | REGION_RW_EL0)
    p.set_exc_guard(GUARD.OFF)

    v1 = p.read32(data_buffer + 8)
    v2 = p.read32(data_buffer + 20)

    print(" Input:", end=" ")
    if v1 == 0:
        print("FLUSH ", end=" ")
    elif v1 == 0x00800000:
        print("NORMAL", end=" ")
    else:
        print("0x08x?" % v1, end=" ")

    print("Output:", end=" ")
    if v2 == 0:
        print("FLUSH ", end=" ")
    elif v2 == 0x00400000:
        print("NORMAL", end=" ")
    else:
        print("0x08x?" % v2, end=" ")
    print("r = 0x%x" % ret)


print("Testing normal mode")
u.msr(ACTLR_EL1, ACTLR_DEFAULT)
u.msr(AFPCR, 0)

u.msr(FPCR, 0)
print("FPCR.FZ = 0")
test_denormals()

u.msr(FPCR, FPCR_FZ)
print("FPCR.FZ = 1")
test_denormals()

print()
print("Testing Apple mode")
u.msr(ACTLR_EL1, ACTLR_DEFAULT | ACTLR_AFP)
u.msr(AFPCR, 0)

u.msr(FPCR, 0)
print("FPCR.FZ = 0")
test_denormals()

u.msr(FPCR, FPCR_FZ)
print("FPCR.FZ = 1")
test_denormals()

u.msr(AFPCR, AFPCR_DAZ)
print("AFPCR.<FTZ, DAZ> = 0, 1")
test_denormals()

u.msr(AFPCR, AFPCR_FTZ)
print("AFPCR.<FTZ, DAZ> = 1, 0")
test_denormals()

u.msr(AFPCR, AFPCR_FTZ | AFPCR_DAZ)
print("AFPCR.<FTZ, DAZ> = 1, 1")
test_denormals()



# SPDX-License-Identifier: MIT
from . import ADTDevTracer
from .dart import DARTTracer
from ..hv import TraceMode
from ..utils import *

from ..hw.isp import *
from ..fw.isp import *
from ..fw.isp.isp_opcodes import *

class ISPCommandDirection(IntEnum):
    RX = 0
    TX = 1

def opcode2name(opcode):
    if opcode_dict and (opcode in opcode_dict): return opcode_dict[opcode]
    return "CISP_CMD_UNK_%04x" % (opcode)

class ISPCommand:
    def __init__(self, chan, msg, direction):
        self.arg0, self.arg1, self.arg2, self.arg3, self.arg4, self.arg5, self.arg6, self.arg7 = struct.unpack("<8q", msg.data)
        self.iova = self.arg0 & ~3
        self.msg = msg
        self.chan = chan
        self.direction = direction
        self.tracer = chan.tracer

    def dump(self):
        self.log(f"[CMD arg0: {hex(self.arg0)}, arg1: {hex(self.arg1)}, arg2: {hex(self.arg2)}]")

    def read_iova(self, iova, size):
        return self.tracer.dart.ioread(0, iova, size)

    def log(self, msg):
        if self.direction == ISPCommandDirection.RX:
            self.tracer.log(f"<== [{self.chan.name}]({self.msg.index}): {msg}")
        else:
            self.tracer.log(f"==> [{self.chan.name}]({self.msg.index}): {msg}")

class ISPTerminalCommand(ISPCommand):
    # Broken as of 13.5.2
    def __init__(self, chan, msg, direction):
        super().__init__(chan, msg, direction)

    def dump(self):
        super().dump()

class ISPIOCommand(ISPCommand):
    def __init__(self, chan, msg, direction):
        super().__init__(chan, msg, direction)
        self.contents = self.read_iova(self.iova, max(self.arg1, 8))

    def dump(self):
        if self.iova:
            opcode = struct.unpack("<Q", self.contents[:0x8])[0] >> 32
            self.log(f"[IO iova: {hex(self.iova)}, insize: {hex(self.arg1)}, outsize: {hex(self.arg2)} -> opcode: {hex(opcode)} {opcode2name(opcode)}]")
            self.log("IO struct: ")
            try:
                chexdump32(self.contents, print_fn=self.log)
            except struct.error:
                chexdump(self.contents, print_fn=self.log)

class ISPT2HBufferCommand(ISPCommand):
    def __init__(self, chan, msg, direction):
        super().__init__(chan, msg, direction)
        self.contents = self.read_iova(self.iova, 0x280)

    def dump(self):
        super().dump()
        if self.contents:
            self.log("BUF_T2H struct:")
            chexdump32(self.contents, print_fn=self.log)

class ISPH2TBufferCommand(ISPCommand):
    def __init__(self, chan, msg, direction):
        super().__init__(chan, msg, direction)
        self.contents = self.read_iova(self.iova, 0x4000)
        #self.contents = None

    def dump(self):
        super().dump()
        if self.contents:
            self.log("BUF_H2T struct:")
            chexdump32(self.contents, print_fn=self.log)

class ISPT2HIOCommand(ISPCommand):
    def __init__(self, chan, msg, direction):
        super().__init__(chan, msg, direction)

    def dump(self):
        super().dump()

class ISPSharedMallocCommand(ISPCommand):
    def __init__(self, chan, msg, direction):
        super().__init__(chan, msg, direction)

    def dump(self):
        if not self.arg0:
            try:
                name = struct.pack(">Q", self.arg2).decode()
            except UnicodeDecodeError:
                name = "UNICODE-GARBAGE-%08x" % (self.arg2)
            self.log("[Malloc REQ: size: 0x%x, name: %s]" % (self.arg1, name))
        else:
            self.log("[Free REQ: iova: 0x%x, index: 0x%x]" % (self.arg0, self.arg2))

class ISPChannelTable:
    def __init__(self, tracer, num_chans, table_iova):
        self.tracer = tracer
        self.table_iova = table_iova
        self.num_chans = num_chans
        self.desc_size = ISPIPCChanTableDescEntry.sizeof()

        chans = []
        table_data = self.tracer.ioread(table_iova, num_chans*self.desc_size)
        for n in range(num_chans):
            entry = table_data[n*self.desc_size:(n+1)*self.desc_size]
            x = ISPIPCChanTableDescEntry.parse(entry)
            chan = ISPTracerChannel(self.tracer, x.name, x.type, x.src, x.num, x.iova)
            chans.append(chan)
        self.chans = chans

    def get_last_rx_commands(self, val):
        for chan in self.chans:
            if (chan.type == 1):
                chan.get_last_commands(ISPCommandDirection.RX)

    def get_last_tx_commands(self, doorbell):
        for chan in self.chans:
            if (chan.doorbell == doorbell):
                chan.get_last_commands(ISPCommandDirection.TX)

class ISPTracerChannel(ISPChannel):
    def __init__(self, isp, name, _type, src, num, iova):
        super().__init__(isp, name, _type, src, num, iova) # init as 'tracer'
        self.tracer = isp

    def __convert2command__(self, msg, direction):
        if self.name == "TERMINAL":
            return ISPTerminalCommand(self, msg, direction)
        elif self.name == "IO" or self.name == "DEBUG":
            return ISPIOCommand(self, msg, direction)
        elif self.name == "SHAREDMALLOC":
            return ISPSharedMallocCommand(self, msg, direction)
        elif self.name == "BUF_T2H":
            return ISPT2HBufferCommand(self, msg, direction)
        elif self.name == "BUF_H2T":
            return ISPH2TBufferCommand(self, msg, direction)
        elif self.name == "IO_T2H":
            return ISPT2HIOCommand(self, msg, direction)
        else:
            return ISPCommand(self, msg, direction)

    def get_last_commands(self, direction):
        cmds = []  # collect asap
        for n in range(self.num):
            pos = (self.cursor + n) % self.num
            dat = self.tracer.ioread(self.iova + (self.entry_size * pos), self.entry_size)
            msg = ISPChannelMessage.parse(dat, index=pos)
            if (not msg.valid()):
                self.cursor = pos
                break
            else:
                cmd = self.__convert2command__(msg, direction)
                cmds.append(cmd)
        for cmd in cmds:
            cmd.dump()

    def __str__(self):
        return f"[{str(self.name)}: src={self.src!s} type={self.type!s} num={self.num!s} iova={hex(self.iova)!s})"

class ISPTracer(ADTDevTracer):

    DEFAULT_MODE = TraceMode.SYNC
    REGMAPS = [ISPRegs]
    NAMES = ["isp"]

    def __init__(self, hv, dev_path, dart_dev_path, verbose):
        super().__init__(hv, dev_path, verbose)
        hv.p.pmgr_adt_clocks_enable(dart_dev_path)
        self.dart_tracer = DARTTracer(hv, dart_dev_path, verbose=0)
        self.dart_tracer.start()
        self.dart = self.dart_tracer.dart
        self.iova_base = 0
        chip_id = hv.adt["/chosen"].chip_id
        if 0x6020 <= chip_id <= 0x6fff:
            self.iova_base = 0x100_0000_0000

        self.ignored_ranges = [
            (0x22c0e8000, 0x4000), # dart 1
            (0x22c0f4000, 0x4000), # dart 2
            (0x22c0fc000, 0x4000), # dart 3
            (0x3860e8000, 0x4000), # dart 1
            (0x3860f4000, 0x4000), # dart 2
            (0x3860fc000, 0x4000), # dart 3
            (0x22c4a8000, 0x4000), # dart 1
            (0x22c4b4000, 0x4000), # dart 2
            (0x22c4bc000, 0x4000), # dart 3
        ]

        self.table = None
        self.num_chans = 0

    def r_ISP_GPIO_0(self, val):
        self.log("ISP_GPIO_0 r32: 0x%x" % (val.value))
        if val.value == 0x8042006:
            self.log(f"ISP_GPIO0 = ACK")
        elif val.value == 0xf7fbdff9:
            self.log(f"ISP_GPIO0 = NACK?")
        elif val.value < 64:
            self.log(f"ISP_IPC_CHANNELS = {val!s}")
            self.num_chans = val.value
        elif val.value > 0:
            self.log(f"IPC BASE IOVA: {val!s}")
            self.ipc_iova = val.value
            # self.dart_tracer.trace_range(0, irange(val.value | self.iova_base, 0x40000))
            self.table = ISPChannelTable(self, self.num_chans, val.value)
            self.log("======== CHANNEL TABLE ========")
            for chan in self.table.chans:
                self.log(f"ISPIPC: {str(chan)}")
            self.log("======== END OF CHANNEL TABLE ========")
    r_ISP_GPIO_0_T8112 = r_ISP_GPIO_0

    def r_ISP_IRQ_INTERRUPT(self, val):
        #self.log("ISP_IRQ_INTERRUPT r32: 0x%x" % (val.value))
        #self.log(f"======== BEGIN IRQ ========")
        self.table.get_last_rx_commands(int(val.value))
        #self.log(f"========  END IRQ  ========")
    r_ISP_IRQ_INTERRUPT_T8112 = r_ISP_IRQ_INTERRUPT

    def w_ISP_IRQ_DOORBELL(self, val):
        #self.log("ISP_IRQ_DOORBELL w32: 0x%x" % (val.value))
        #self.log(f"======== BEGIN DOORBELL ========")
        self.table.get_last_tx_commands(int(val.value))
        #self.log(f"========  END DOORBELL  ========")
    w_ISP_IRQ_DOORBELL_T8112 = w_ISP_IRQ_DOORBELL

    def w_ISP_GPIO_0(self, val):
        self.log("ISP_GPIO_0 w32: 0x%x" % (val.value))
        if (val.value >= 0xe00000) and (val.value <= 0x1100000): # dunno
            self.log("ISP bootargs at 0x%x:" % val.value)
            bootargs = self.dart.ioread(0, val.value | self.iova_base, 0x200) # justt in case
            chexdump32(bootargs, print_fn=self.log)
            x = ISPIPCBootArgs.parse(bootargs[:ISPIPCBootArgs.sizeof()])
            self.log(x)
    w_ISP_GPIO_0_T8112 = w_ISP_GPIO_0

    def ioread(self, iova, size):
        iova |= self.iova_base
        return self.dart.ioread(0, iova, size)

    def iowrite(self, iova, data):
        iova |= self.iova_base
        return self.dart.iowrite(0, iova, data)

    def start(self):
        super().start()
        for addr, size in self.ignored_ranges:
            self.trace(addr, size, TraceMode.OFF)



# SPDX-License-Identifier: MIT
import atexit, serial, os, struct, code, traceback, readline, rlcompleter, sys
import __main__
import builtins
import re

from .proxy import *
from .proxyutils import *
from .utils import *
from . import sysreg
from inspect import isfunction, signature

__all__ = ["ExitConsole", "run_shell"]

class HistoryConsole(code.InteractiveConsole):
    def __init__(self, locals=None, filename="<console>",
                 histfile=os.path.expanduser("~/.m1n1-history")):
        code.InteractiveConsole.__init__(self, locals, filename)
        self.histfile = histfile
        self.init_history(histfile)
        self.poll_func = None

    def init_history(self, histfile):
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
            except FileNotFoundError:
                pass

    def save_history(self):
        readline.set_history_length(10000)
        try:
            readline.write_history_file(self.histfile)
        except OSError as e:
            print(f"Failed writing history to {self.histfile}: {e}", file=sys.stderr)

    def showtraceback(self):
        type, value, tb = sys.exc_info()
        traceback.print_exception(type, value, tb)

    def runcode(self, code):
        super().runcode(code)
        if self.poll_func:
            self.poll_func()
        if "mon" in self.locals:
            try:
                self.locals["mon"].poll()
            except Exception as e:
                print(f"mon.poll() failed: {e!r}")
        if "u" in self.locals:
            self.locals["u"].push_simd()


class ExitConsole(SystemExit):
    pass
cmd_list = {}
subcmd_list = {}
# Debug levels
DBL_NONE = 0
DBL_INFO = 1
DBL_TRACE = 2
DBL_DEBUG = 3
DBL_EDEBUG = 4

db_level = DBL_NONE

def debug_cmd(db=None):
    '''Set debug level to integer %d(none)...%d(extreme debug)''' % (DBL_NONE, DBL_EDEBUG)
    global db_level
    if db:
        db_level = db
    print("debug level=%d" % db_level)

def help_cmd(arg=None):
    if db_level >= DBL_DEBUG:
        print("arg=%s" % repr(arg))
    if arg:
        #cmd = arg.__qualname__
        if callable(arg):
            cmd = arg.__name__
        elif isinstance(arg, str):
            cmd = arg
        else:
            print("Unknown command: %s" % repr(arg))
            return
        if db_level >= DBL_DEBUG:
            print("cmd=%s" % repr(cmd))
        if cmd not in cmd_list:
            print("Undocumented command %s" % cmd)
            return
        hinfo = cmd_list[cmd]
        if isinstance(hinfo, str):
            print("%-10s : %s" % (cmd, hinfo))
            return
        if cmd in subcmd_list:
            clist = subcmd_list[cmd]
            aname = cmd
            if db_level >= DBL_DEBUG:
                print("subcmd_list[%s] = %s" %
                    (repr(cmd), repr(clist)))
        else:
            print("command %s is not documented" % cmd)
            return
    else:
        clist = cmd_list
        aname = 'top level'
        print("Note: To display a category's commands quote the name e.g. help('HV')")
    print("List of %s commands:" % aname)
    for cmd in clist.keys():
        hinfo = clist[cmd]
        if isinstance(hinfo, str):
            msg = hinfo.strip().split('\n', 1)[0]
        elif isinstance(hinfo, int):
            msg = "%s category - %d subcommands" % (cmd, hinfo)
        else:
            print("%s ?" % cmd)
            continue
        if len(cmd) <= 10:
            print("%-10s : %s" % (cmd, msg))
        else:
            print("%s:\n             %s" % (cmd, msg))

#locals is a dictionary for constructing the
# InteractiveConsole with. It adds in the callables
# in proxy utils iface and sysreg into locals
def run_shell(locals, msg=None, exitmsg=None, poll_func=None):
    saved_display = sys.displayhook
    try:
        def display(val):
            if isinstance(val, int) and not isinstance(val, bool):
                builtins._ = val
                print(hex(val))
            elif callable(val):
                val()
            else:
                saved_display(val)

        sys.displayhook = display

        # convenience
        locals["h"] = hex
        locals["sysreg"] = sysreg

        if "proxy" in locals and "p" not in locals:
            locals["p"] = locals["proxy"]
        if "utils" in locals and "u" not in locals:
            locals["u"] = locals["utils"]

        for obj_name in ("iface", "p", "u"):
            obj = locals.get(obj_name)
            obj_class = type(obj)
            if obj is None:
                continue

            for attr in dir(obj_class):
                if attr in locals or attr.startswith('_'):
                    continue

                member = getattr(obj_class, attr)
                if callable(member) and not isinstance(member, property):
                    cmd = getattr(obj, attr)
                    locals[attr] = cmd

        for attr in dir(sysreg):
            locals[attr] = getattr(sysreg, attr)

        locals['help'] = help_cmd
        locals['debug'] = debug_cmd
        for obj_name in locals.keys():
            obj = locals.get(obj_name)
            if obj is None or obj_name.startswith('_'):
                continue
            if callable(obj) and not isinstance(obj, property):
                try:
                    desc = obj_name + str(signature(obj))
                except:
                    continue
                qn = obj.__qualname__
                if qn.find('.') > 0:
                    a = qn.split('.')
                    if a[0] not in subcmd_list:
                        subcmd_list[a[0]] = {}
                    if a[0] not in cmd_list:
                        cmd_list[a[0]] = 1
                    else:
                        cmd_list[a[0]] += 1
                    clist = subcmd_list[a[0]] 
                else:
                    clist = None
                if locals[obj_name].__doc__:
                    desc += " - " + locals[obj_name].__doc__
                    cmd_list[obj_name] = desc
                if isinstance(clist, dict):
                    clist[obj_name] = desc

        con = HistoryConsole(locals)
        con.poll_func = poll_func
        try:
            con.interact(msg, exitmsg)
        except ExitConsole as e:
            if len(e.args):
                return e.args[0]
            else:
                return
        finally:
            con.save_history()

    finally:
        sys.displayhook = saved_display

if __name__ == "__main__":
    from .setup import *
    locals = dict(__main__.__dict__)

    run_shell(locals, msg="Have fun!")



# SPDX-License-Identifier: MIT
from ..common import Padding
from construct import *
import cv2
import datetime
import numpy as np
import struct
import _thread

from .isp_cmd import ISPIOCommandDispatcher

ISPFrameMeta = Struct(
        "unk_0" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),
        "unk_8" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),

        "meta_iova" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),
        "pad" / Default(Int32ul, 0),
        "pad" / Default(Int32ul, 0),

        "pad" / Padding(0x10),

        "unk_30" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),
        "pad" / Default(Int32ul, 0),
        "pad" / Default(Int32ul, 0),

        "unk_40" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),
        "unk_48" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),

        "luma_iova" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),
        "cbcr_iova" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),

        "pad" / Padding(0x20),

        "unk_80" / Hex(Int32ul),
        "unk_84" / Hex(Int32ul),
        "unk_88" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),

        "pad" / Padding(0x30),
    "pad" / Padding(0x140),

    "pad" / Padding(0x10),

        "unk_210" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),
        "index" / Hex(Int32ul),
        "unk_21c" / Hex(Int32ul),

        "unk_220" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),
        "pad" / Default(Int32ul, 0),
        "pad" / Default(Int32ul, 0),

        "unk_230" / Hex(Int32ul),
        "unk_234" / Hex(Int32ul),
        "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),

        "pad" / Padding(0x40),
)
assert((ISPFrameMeta.sizeof() == 0x280))


class ISPFrame:
    def __init__(self, isp, req):
        self.isp = isp
        self.height = 720
        self.width = 1280
        self.meta_size = 0x4640
        self.luma_size = self.height * self.width       # 1280 * 720; 921600; 0xe1000
        self.cbcr_size = self.height * self.width // 2  # 1280 * 360; 460800; 0x70800

        assert((req.arg1 == 0x280))
        x = ISPFrameMeta.parse(self.isp.ioread(req.arg0, req.arg1))
        self.meta_iova = x.meta_iova
        self.luma_iova = x.luma_iova
        self.cbcr_iova = x.cbcr_iova
        self.index = x.index

        self.luma_data = self.isp.ioread(self.luma_iova, self.luma_size)
        self.cbcr_data = self.isp.ioread(self.cbcr_iova, self.cbcr_size)

        self.timestamp = datetime.datetime.now()

    def to_bgr(self):  # TODO
        y = np.frombuffer(self.luma_data[:1280*360*2], dtype=np.uint8).reshape((720, 1280))
        cbcr = np.frombuffer(self.cbcr_data[:1280*360*1], dtype=np.uint8).reshape((360, 1280))
        u = cv2.resize(cbcr[:,::2], (1280, 720))
        v = cv2.resize(cbcr[:,1::2], (1280, 720))
        yuv = np.stack((y, u, v), axis=-1)
        bgr = cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR)
        return bgr

    def process(self):
        bgr = self.to_bgr()
        mirrored = bgr[:,::-1,:].copy()
        s = "Frame %d: %s" % (self.index, str(self.timestamp))
        cv2.putText(mirrored, s, (10, self.height - 20), cv2.FONT_HERSHEY_DUPLEX, 0.7, (0,255,0), 1, cv2.LINE_AA)
        return mirrored

    def __str__(self):
        s = "Frame %d: [luma: 0x%x cbcr: 0x%x] at %s" % (self.index, self.luma_iova, self.cbcr_iova, self.timestamp.strftime('%H:%M:%S.%f'))
        return s


BufH2TSendArgsHeader = Struct(
    "unk_0" / Int32ul,
    "pad" / Default(Int32ul, 0),
        "batch" / Int32ul,
    "pad" / Default(Int32ul, 0),
)

BufH2TSendArgs = Struct(
    "iova0" / Int32ul,
    "pad" / Default(Int32ul, 0),
        "iova1" / Int32ul,
    "pad" / Default(Int32ul, 0),
        "pad" / Padding(0x10),
    "flag0" / Int32ul,
        "flag1" / Int32ul,
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "unk_30" / Int32ul,
        "pool" / Int32ul,
    "tag" / Int32ul,
    "pad" / Default(Int32ul, 0),
)
assert((BufH2TSendArgs.sizeof() == 0x40))


class ISPBufH2TBuffer:
    def __init__(self, surf0, surf1, buftype, index, args):
        self.surf0 = surf0
        self.surf1 = surf1
        self.buftype = buftype
        self.index = index
        self.args = args


class ISPBufH2TPool:
    def __init__(self, isp, bufs, header):
        self.isp = isp
        self.bufs = bufs
        self.args = header + b''.join([buf.args for buf in self.bufs])

    @classmethod
    def meta_pool(cls, isp, batch=2):
        bufs = []
        for n in range(10):
            bufs.append(cls.make_metabuf(isp, n))
        header = BufH2TSendArgsHeader.build(dict(
            unk_0=0x1,
            batch=batch,
        ))
        return cls(isp, bufs, header)

    @classmethod
    def yuv_pool(cls, isp, batch=2):
        bufs = []
        for n in range(batch):
            bufs.append(cls.make_yuvbuf(isp, n))
        for n in range(10-batch):
            bufs.append(cls.make_unkbuf(isp, n+batch))

        header = BufH2TSendArgsHeader.build(dict(
            unk_0=0x1,
            batch=batch,
        ))
        return cls(isp, bufs, header)

    @staticmethod
    def make_metabuf(isp, index):
        # CImageCaptureCore.cpp, 4788: FrontRGB: Cannot get host meta buffer, dropping frame #2
        # (461): ISPASC: WRN: ./h10isp/filters/IC/CImageCaptureCore.cpp, 4997: FrontRGB: Host metadata unavailable.
        # AppleH13CamIn::ISP_SendBuffers_gated - h2tBuf: pool=0, tag=0x13f, addr=0x056A8000, len=17984
        # 056a8000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
        # 40000000 00000000 00000000 00000000 00000001 00000000 0000013f 00000000
        surf = isp.mmger.alloc_size(0x4640, name="META")  # 17984
        args = BufH2TSendArgs.build(dict(
                iova0=surf.iova,
            iova1=0x0,
                flag0=0x40000000,
            flag1=0x0,
            unk_30=0x1,
            pool=0x0,
            tag=surf.index,
        ))
        return ISPBufH2TBuffer(surf0=surf, surf1=None, buftype=0, index=index, args=args)

    @staticmethod
    def make_yuvbuf(isp, index):
        # FLOW ERR: ./h10isp/filters/Flow/H13/CVideoFlowH13.cpp, 5351: flow=0, ch=0 can't get output yuv buf, dropped frame=0 (1) 0
        # AppleH13CamIn::ISP_SendBuffers_gated - h2tBuf: pool=1, tag=0x16e, addr0=0x0754C040, len0=921600, addr1=0x0762D040, len1=460800
        # 0754c040 00000000 0762d040 00000000 00000000 00000000 00000000 00000000
        # 40000000 40000000 00000000 00000000 00000002 00000001 0000016e 00000000
        surf0 = isp.mmger.alloc_size(0xe1000, name="LUMA")  # 921600; 1280 * 720
        surf1 = isp.mmger.alloc_size(0x70800, name="CBCR")  # 460800; 1280 * 360
        args = BufH2TSendArgs.build(dict(
                iova0=surf0.iova,
            iova1=surf1.iova,
                flag0=0x40000000,
            flag1=0x40000000,
            unk_30=0x2,
            pool=0x1,
            tag=surf0.index,
        ))
        return ISPBufH2TBuffer(surf0=surf0, surf1=surf1, buftype=1, index=index, args=args)

    @staticmethod
    def make_unkbuf(isp, index):
        # AppleH13CamIn::ISP_SendBuffers_gated - h2tBuf: pool=2, tag=0x165, addr=0x074F8000, len=16960
        # 074f8000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
        # 40000000 00000000 00000000 00000000 00000001 00000002 00000165 00000000
        surf = isp.mmger.alloc_size(0x4240, name="UNK")  # 16960
        args = BufH2TSendArgs.build(dict(
                iova0=surf.iova,
            iova1=0x0,
                flag0=0x40000000,
            flag1=0x0,
            unk_30=0x1,
            pool=0x2,
            tag=surf.index,
        ))
        return ISPBufH2TBuffer(surf0=surf, surf1=None, buftype=2, index=index, args=args)

    def send(self):
        # TX: REQ: [0x1813140, 0x280, 0x30000000]  # iova, size, flag
        # RX: RSP: [0x1813141, 0x000, 0x80000000]  # iova, zero, flag
        req = ISPChannelMessage.build(
            arg0 = self.isp.cmd_iova,  # iova
            arg1 = 0x280,  # size
            arg2 = 0x30000000,  # FFW_INTERPROC_BUFF_EXCHANGE_FLAG_CHECK
        )
        # print("CHAN: BR: TX: REQ: [iova: 0x%x, size: 0x%x, flag: 0x%x]" % (req.arg0, req.arg1, req.arg2))
        self.isp.iowrite(req.arg0, self.args)
        rsp = self.isp.table.bufh2t.send(req)
        if (rsp == None):
            self.isp.table.dump()
            raise RuntimeError("failed to send buf")
        return rsp


class ISPFrameReceiver:
    def __init__(self, isp, batch=2):
        assert((batch <= 10))
        self.isp = isp
        self.meta_pool = ISPBufH2TPool.meta_pool(self.isp, batch)
        self.yuv_pool = ISPBufH2TPool.yuv_pool(self.isp, batch)
        self.dp = ISPIOCommandDispatcher(self.isp)

    def work(self):
        self.yuv_pool.send()
        self.isp.table.buft2h.handler()
        self.meta_pool.send()
        self.isp.table.buft2h.handler()

    @staticmethod
    def input_thread(a_list):  # https://stackoverflow.com/a/25442391
        input()
        a_list.append(True)

    def work_loop(self):
        # self.meta_pool.send()  # sent ahead of time
        a_list = []
        _thread.start_new_thread(self.input_thread, (a_list,))
        while not a_list:
            self.work()

    def stream(self):
        print("Kicking up hardware...")
        self.ch_start()
        print("Starting stream...")
        self.work_loop()
        cv2.destroyAllWindows()
        print("Ended stream.")
        self.ch_stop()
        print("Powered down hardware.")

    def ch_start(self):
        dp = self.dp
        dp.cmd_print_enable()
        dp.cmd_trace_enable()

        dp.cmd_set_isp_pmu_base()
        dp.cmd_set_dsid_clr_req_base2()
        dp.cmd_pmp_ctrl_set()
        dp.cmd_start()

        dp.cmd_ch_camera_config_select()
        dp.cmd_ch_sbs_enable()

        dp.cmd_ch_crop_set()
        dp.cmd_ch_output_config_set()
        dp.cmd_ch_preview_stream_set()

        dp.cmd_ch_cnr_start()
        dp.cmd_ch_mbnr_enable()

        dp.cmd_apple_ch_temporal_filter_start()
        dp.cmd_apple_ch_motion_history_start()
        dp.cmd_apple_ch_temporal_filter_enable()
        dp.cmd_apple_ch_ae_fd_scene_metering_config_set()
        dp.cmd_apple_ch_ae_metering_mode_set()
        dp.cmd_ch_ae_stability_set()
        dp.cmd_ch_ae_stability_to_stable_set()

        dp.cmd_ch_sif_pixel_format_set()
        #dp.cmd_ch_buffer_recycle_mode_set()
        #dp.cmd_ch_buffer_recycle_start()

        dp.cmd_ch_ae_frame_rate_max_set()
        dp.cmd_ch_ae_frame_rate_min_set()
        dp.cmd_ch_buffer_pool_config_set()

        # call right before cmd_ch_start() as to not drop start frames
        self.meta_pool.send()
        dp.cmd_ch_start()

    def ch_stop(self):
        dp = self.dp
        dp.cmd_ch_buffer_recycle_stop()
        dp.cmd_ch_stop()
        dp.cmd_stop()



# SPDX-License-Identifier: MIT
from ..utils import *

class R_PSTATE(Register32):
    RESET = 31
    AUTO_ENABLE = 28
    AUTO_STATE = 27, 24
    PARENT_MISSING = 11
    DEV_DISABLE = 10
    WAS_CLKGATED = 9
    WAS_PWRGATED = 8
    ACTUAL = 7, 4
    DESIRED = 3, 0

class R_PWRGATE(Register32):
    GATE = 31

class R_CLK_CFG(Register32):
    UNK31 = 31
    SRC = 30, 24
    UNK20 = 20
    UNK8 = 8
    UNK0 = 7, 0

class PMGRRegs0(RegMap):
    PS3 = irange(0x0000, 10, 8), R_PSTATE
    PS4 = irange(0x0200, 32, 8), R_PSTATE
    PS5 = irange(0x0300, 32, 8), R_PSTATE
    PS6 = irange(0x0c00, 2, 8), R_PSTATE
    PS7 = irange(0x4000, 13, 8), R_PSTATE
    PS8 = irange(0x8000, 5, 8), R_PSTATE
    PS9 = irange(0xc000, 7, 8), R_PSTATE
    PS10 = irange(0x10000, 10, 8), R_PSTATE
    PS11 = irange(0x100, 32, 8), R_PSTATE
    PS12 = irange(0x400, 15, 8), R_PSTATE

    PG1 = irange(0x1c010, 16, 8), R_PWRGATE
    PG1CFG = (irange(0x1c090, 69, 24), irange(0, 6, 4)), Register32

    CPUTVM0 = 0x48000, Register32
    CPUTVM1 = 0x48c00, Register32
    CPUTVM2 = 0x48800, Register32
    CPUTVM3 = 0x48400, Register32

class PMGRRegs1(RegMap):
    PS0 = irange(0x58, 32, 8), R_PSTATE
    PS1 = irange(0x4000, 32, 8), R_PSTATE
    PS2 = irange(0x8000, 32, 8), R_PSTATE

    PG0 = irange(0x1c010, 32, 8), R_PWRGATE

class PMGRRegs2(RegMap):
    CLK_CFG0 = irange(0x40000, 86, 4), R_CLK_CFG
    CLK_CFG1 = irange(0x40200, 8, 4), R_CLK_CFG
    CLK_CFG2 = irange(0x40280, 2, 4), R_CLK_CFG

class PMGR:
    def __init__(self, u):
        self.u = u
        self.p = u.proxy
        self.iface = u.iface
        self.node = u.adt["/arm-io/pmgr"]
        self.regs = [
            PMGRRegs0(u, self.node.get_reg(0)[0]),
            PMGRRegs1(u, self.node.get_reg(1)[0]),
            PMGRRegs2(u, self.node.get_reg(2)[0]),
        ]

    def dump_all(self):
        for i in self.regs:
            i.dump_regs()



# SPDX-License-Identifier: MIT
from ..utils import *
from .dart import DARTRegs
import time


class ANERegs(RegMap):
	PMGR1 = 0x738, Register32
	PMGR2 = 0x798, Register32
	PMGR3 = 0x7f8, Register32

	ASC_IO_RVBAR = 0x1050000, Register32
	ASC_EDPRCR   = 0x1010310, Register32

	# 24hz clocks, counter at +4
	CLK0 = 0x1160008, Register32
	CTR0 = 0x116000c, Register32
	CLK1 = 0x1168008, Register32
	CTR1 = 0x116800c, Register32
	CLK2 = 0x1170000, Register32
	CTR2 = 0x1170004, Register32
	CLK3 = 0x1178000, Register32
	CTR3 = 0x1178004, Register32

	VERS = 0x1840000, Register32

	# for acks w/ rtkit
	GPIO0 = 0x1840048, Register32
	GPIO1 = 0x184004c, Register32
	GPIO2 = 0x1840050, Register32
	GPIO3 = 0x1840054, Register32
	GPIO4 = 0x1840058, Register32
	GPIO5 = 0x184005c, Register32
	GPIO6 = 0x1840060, Register32
	GPIO7 = 0x1840064, Register32


class ANEDARTRegs(DARTRegs):
	UNK_CONFIG_68 = 0x68, Register32
	UNK_CONFIG_6c = 0x6c, Register32


class R_TQINFO(Register32):
	UNK = 31, 16
	NID = 15, 0

class TaskQueue(RegMap):
	STATUS     = irange(0x00, 8, 0x148), Register32
	PRTY       = irange(0x10, 8, 0x148), Register32
	FREE_SPACE = irange(0x14, 8, 0x148), Register32
	TQINFO     = irange(0x1c, 8, 0x148), R_TQINFO

	BAR1 = (irange(0x20, 8, 0x148), irange(0x0, 0x20, 4)), Register32
	REQ_NID1  = irange(0xa0, 8, 0x148), Register32
	REQ_SIZE2 = irange(0xa4, 8, 0x148), Register32
	REQ_ADDR2 = irange(0xa8, 8, 0x148), Register32

	BAR2 = (irange(0xac, 8, 0x148), irange(0x0, 0x20, 4)), Register32
	REQ_NID2  = irange(0x12c, 8, 0x148), Register32
	REQ_SIZE1 = irange(0x130, 8, 0x148), Register32
	REQ_ADDR1 = irange(0x134, 8, 0x148), Register32


class R_REQINFO(Register32):
	TDSIZE  = 31, 16
	TDCOUNT = 15,  0

class R_IRQINFO(Register32):
	CNT  = 31, 24
	NID  = 23, 16
	UNK1 = 15, 8
	UNK2 = 7,  0

class TMRegs(RegMap):
	REQ_ADDR = 0x0, Register32
	REQ_INFO = 0x4, R_REQINFO
	REQ_PUSH = 0x8, Register32
	TQ_EN    = 0xc, Register32

	IRQ_EVT1_CNT      = 0x14, Register32
	IRQ_EVT1_DAT_INFO = 0x18, R_IRQINFO
	IRQ_EVT1_DAT_UNK1 = 0x1c, Register32
	IRQ_EVT1_DAT_TIME = 0x20, Register32
	IRQ_EVT1_DAT_UNK2 = 0x24, Register32

	IRQ_EVT2_CNT      = 0x28, Register32
	IRQ_EVT2_DAT_INFO = 0x2c, R_IRQINFO
	IRQ_EVT2_DAT_UNK1 = 0x30, Register32
	IRQ_EVT2_DAT_TIME = 0x34, Register32
	IRQ_EVT2_DAT_UNK2 = 0x38, Register32

	COMMIT_INFO = 0x44, Register32
	TM_STATUS   = 0x54, Register32

	UNK_IRQ_EN1 = 0x68, Register32
	UNK_IRQ_ACK = 0x6c, Register32
	UNK_IRQ_EN2 = 0x70, Register32


class ANETaskManager:

	TQ_COUNT = 8
	TQ_WIDTH = 0x148
	tq_prty = (0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x1e, 0x1f)

	def __init__(self, ane):
		self.u = ane.u
		self.p = ane.p
		self.TM_BASE_ADDR = ane.base_addr + 0x1c00000 + 0x24000
		self.TQ_BASE_ADDR = ane.base_addr + 0x1c00000 + 0x25000
		self.regs = TMRegs(self.u, self.TM_BASE_ADDR)
		self.tq = TaskQueue(self.u, self.TQ_BASE_ADDR)

	def reset(self):  # these reset with ANE_SET pds
		self.regs.TQ_EN.val = 0x3000

		# set priority param for each queue
		for qid, prty in enumerate(self.tq_prty):
			self.tq.PRTY[qid].val = self.tq_prty[qid]

		self.regs.UNK_IRQ_EN1.val = 0x4000000  # enable irq
		self.regs.UNK_IRQ_EN2.val = 0x6  # enable irq

	def enqueue_tq(self, req):
		qid = req.qid
		if not ((qid >= 1) and (qid < self.TQ_COUNT)):
			raise ValueError('1 <= qid <= 7')
		if not (self.tq.PRTY[qid].val == self.tq_prty[qid]):
			raise ValueError('invalid priority setup for tq %d' % qid)

		print('enqueueing task w/ fifo 0x%x to tq %d' % (req.fifo_iova, qid))
		self.tq.STATUS[qid].val = 0x1  # in use

		for bdx, iova in enumerate(req.bar):
			if (iova):
				print("bar %d: 0x%x" % (bdx, iova))
			self.tq.BAR1[qid, bdx].val = iova

		self.tq.REQ_SIZE1[qid].val = ((req.td_size << 0xe) + 0x1ff0000) & 0x1ff0000
		self.tq.REQ_ADDR1[qid].val = req.fifo_iova & 0xffffffff
		self.tq.REQ_NID1[qid].val = (req.nid & 0xff) << 8 | 1

	def execute_tq(self, req):
		qid = req.qid
		print('arbitered tq %d; pushing to execution queue...' % qid)

		# transfer to main queue (now in in TM range)
		self.regs.REQ_ADDR.val = self.tq.REQ_ADDR1[qid].val
		# doesnt go through if 0
		self.regs.REQ_INFO.val = self.tq.REQ_SIZE1[qid].val | req.td_count
		# let's do magic
		self.regs.REQ_PUSH.val = self.tq_prty[qid] | (qid & 7) << 8

		self.get_tm_status()
		self.get_committed_info()
		self.irq_handler()
		self.tq.STATUS[qid].val = 0x0  # done

	def get_tm_status(self, max_timeouts=100, interval=0.01):
		for n in range(max_timeouts):
			status = self.regs.TM_STATUS.val
			success = (status & 1) != 0
			print('tm status: 0x%x, success: %r' % (status, success))
			if (success):
				return success
			time.sleep(interval)
		print('timeout, tm is non-idle! status: 0x%x' % status)
		return success

	def get_committed_info(self):
		committed_nid = self.regs.COMMIT_INFO.val >> 0x10 & 0xff
		print('pushed td w/ nid 0x%x to execution' % committed_nid)

	def irq_handler(self):
		line = 0
		evtcnt = self.regs.IRQ_EVT1_CNT.val
		print('irq handler: LINE %d EVTCNT: %d' % (line, evtcnt))
		for evt_n in range(evtcnt):  # needs to be cleared
			info = self.regs.IRQ_EVT1_DAT_INFO.val
			unk1 = self.regs.IRQ_EVT1_DAT_UNK1.val
			tmstmp = self.regs.IRQ_EVT1_DAT_TIME.val
			unk2 = self.regs.IRQ_EVT1_DAT_UNK2.val
			print('irq handler: LINE %d EVT %d: executed info 0x%x @ 0x%x'
				  % (line, evt_n, info, tmstmp))

		self.regs.UNK_IRQ_ACK.val = self.regs.UNK_IRQ_ACK.val | 2

		line = 1
		evtcnt = self.regs.IRQ_EVT2_CNT.val
		print('irq handler: LINE %d EVTCNT: %d' % (line, evtcnt))
		for evt_n in range(evtcnt):  # needs to be cleared
			info = self.regs.IRQ_EVT2_DAT_INFO.val
			unk1 = self.regs.IRQ_EVT2_DAT_UNK1.val
			tmstmp = self.regs.IRQ_EVT2_DAT_TIME.val
			unk2 = self.regs.IRQ_EVT2_DAT_UNK2.val
			print('irq handler: LINE %d EVT %d: executed info 0x%x @ 0x%x'
				  % (line, evt_n, info, tmstmp))



# SPDX-License-Identifier: MIT
import sys, time
from enum import IntEnum
from ..utils import *

__all__ = ["ADMACRegs", "ADMAC", "E_BUSWIDTH", "E_FRAME"]


class R_RING(Register32):
    # overflow/underflow counter
    OF_UF = 31, 16

    # goes through 0, 1, 2, 3 as the pieces of a report/descriptor
    # are being read/written through REPORT_READ/DESC_WRITE
    READOUT_PROGRESS = 13, 12

    # when READ_SLOT==WRITE_SLOT one of the two is set
    EMPTY = 8
    FULL = 9

    ERR = 10

    # next slot to read
    READ_SLOT = 5, 4

    # next slot to be written to
    WRITE_SLOT = 1, 0

class R_CHAN_STATUS(Register32):
    # only raised if the descriptor had NOTIFY set
    DESC_DONE = 0

    DESC_RING_EMPTY = 4
    REPORT_RING_FULL = 5

    # cleared by writing ERR=1 either to TX_DESC_RING or TX_REPORT_RING
    RING_ERR = 6

    UNK0 = 1
    UNK3 = 8
    UNK4 = 9
    UNK5 = 10

class R_CHAN_CONTROL(Register32):
    RESET_RINGS = 0
    CLEAR_OF_UF_COUNTERS = 1
    UNK1 = 3

class E_BUSWIDTH(IntEnum):
    W_8BIT  = 0
    W_16BIT = 1
    W_32BIT = 2

class E_FRAME(IntEnum):
    F_1_WORD  = 0
    F_2_WORDS = 1
    F_4_WORDS = 2

class R_BUSWIDTH(Register32):
    WORD  = 2, 0, E_BUSWIDTH
    FRAME = 6, 4, E_FRAME

class R_CARVEOUT(Register32):
    SIZE = 31, 16
    BASE = 15, 0

class ADMACRegs(RegMap):
    TX_EN     = 0x0, Register32 # one bit per channel
    TX_EN_CLR = 0x4, Register32

    RX_EN     = 0x8, Register32
    RX_EN_CLR = 0xc, Register32

    UNK_CTL = 0x10, Register32

    # each of the four registers represents an internal interrupt line,
    # bits represent DMA channels which at the moment raise that particular line
    #
    # the irq-destination-index prop in ADT maybe selects the line which
    # is actually wired out
    #
    TX_INTSTATE = irange(0x30, 4, 0x4), Register32
    RX_INTSTATE = irange(0x40, 4, 0x4), Register32

    # a 24 MHz always-running counter, top bit is always set
    COUNTER = 0x70, Register64

    TX_SRAM_SIZE = 0x94, Register32
    RX_SRAM_SIZE = 0x98, Register32

    # -- per-channel registers --

    CHAN_CTL = (irange(0x8000, 32, 0x200)), R_CHAN_CONTROL

    CHAN_BUSWIDTH      = (irange(0x8040, 32, 0x200)), R_BUSWIDTH
    CHAN_SRAM_CARVEOUT = (irange(0x8050, 32, 0x200)), R_CARVEOUT
    CHAN_BURSTSIZE     = (irange(0x8054, 32, 0x200)), Register32

    CHAN_RESIDUE = irange(0x8064, 32, 0x200), Register32

    CHAN_DESC_RING   = irange(0x8070, 32, 0x200), R_RING
    CHAN_REPORT_RING = irange(0x8074, 32, 0x200), R_RING

    TX_DESC_WRITE  = irange(0x10000, 16, 4), Register32
    TX_REPORT_READ = irange(0x10100, 16, 4), Register32

    RX_DESC_WRITE  = irange(0x14000, 16, 4), Register32
    RX_REPORT_READ = irange(0x14100, 16, 4), Register32

    # per-channel, per-internal-line
    CHAN_STATUS  = (irange(0x8010, 32, 0x200), irange(0x0, 4, 0x4)), R_CHAN_STATUS
    CHAN_INTMASK = (irange(0x8020, 32, 0x200), irange(0x0, 4, 0x4)), R_CHAN_STATUS


class ADMACDescriptorFlags(Register32):
    # whether to raise DESC_DONE in CHAN_STATUS
    NOTIFY = 16

    # whether to repeat this descriptor ad infinitum
    #
    # once a descriptor with this flag is loaded, any descriptors loaded
    # afterwards are also repeated and nothing short of full power domain reset
    # seems to revoke that behaviour. this looks like a HW bug.
    REPEAT = 17

    # arbitrary ID propagated into reports
    DESC_ID = 7, 0

class ADMACDescriptor(Reloadable):
    def __init__(self, addr, length, **flags):
        self.addr = addr
        self.length = length
        self.flags = ADMACDescriptorFlags(**flags)

    def __repr__(self):
        return f"<descriptor: addr=0x{self.addr:x} len=0x{self.length:x} flags={self.flags}>"

    def ser(self):
        return [
            self.addr & (1<<32)-1,
            self.addr>>32 & (1<<32)-1,
            self.length & (1<<32)-1,
            int(self.flags)
        ]

    @classmethod
    def deser(self, seq):
        if not len(seq) == 4:
            raise ValueError
        return ADMACDescriptor(
            seq[0] | seq[1] << 32, # addr
            seq[2], # length (in bytes)
            **ADMACDescriptorFlags(seq[3]).fields
        )


class ADMACReportFlags(Register32):
    UNK1 = 24
    UNK2 = 25
    UNK4 = 26 # memory access fault?
    UNK3 = 27
    DESC_ID = 7, 0

class ADMACReport(Reloadable):
    def __init__(self, countval, unk1, flags):
        self.countval, self.unk1, self.flags = countval, unk1, ADMACReportFlags(flags)

    def __repr__(self):
        return f"<report: countval=0x{self.countval:x} unk1=0x{self.unk1:x} flags={self.flags}>"

    def ser(self):
        return [
            self.countval & (1<<32)-1,
            self.countval>>32 & (1<<32)-1,
            self.unk1 & (1<<32)-1,
            int(self.flags)
        ]

    @classmethod
    def deser(self, seq):
        if not len(seq) == 4:
            raise ValueError
        return ADMACReport(
            seq[0] | seq[1] << 32, # countval
            seq[2], # unk1
            seq[3] # flags
        )


class ADMACChannel(Reloadable):
    def __init__(self, parent, channo):
        self.p = parent
        self.iface = parent.p.iface
        self.dart = parent.dart
        self.regs = parent.regs
        self.tx = (channo % 2) == 0
        self.rx = not self.tx
        self.ch = channo

        self._desc_id = 0
        self._submitted = {}
        self._last_report = None
        self._est_byte_rate = None

    def reset(self):
        self.regs.CHAN_CTL[self.ch].set(RESET_RINGS=1, CLEAR_OF_UF_COUNTERS=1)
        self.regs.CHAN_CTL[self.ch].set(RESET_RINGS=0, CLEAR_OF_UF_COUNTERS=0)

        self.burstsize = 0xc0_0060
        self.buswidth = E_BUSWIDTH.W_32BIT
        self.framesize = E_FRAME.F_1_WORD

    def enable(self):
        self.regs.CHAN_INTMASK[self.ch, 0].reg = \
                R_CHAN_STATUS(DESC_DONE=1, DESC_RING_EMPTY=1,
                                REPORT_RING_FULL=1, RING_ERR=1)

        if self.tx:
            self.regs.TX_EN.val = 1 << (self.ch//2)
        else:
            self.regs.RX_EN.val = 1 << (self.ch//2)

    def disable(self):
        if self.tx:
            self.regs.TX_EN_CLR.val = 1 << (self.ch//2)
        else:
            self.regs.RX_EN_CLR.val = 1 << (self.ch//2)

    @property
    def buswidth(self):
        self.regs.CHAN_BUSWIDTH[self.ch].reg.WORD

    @buswidth.setter
    def buswidth(self, wordsize):
        return self.regs.CHAN_BUSWIDTH[self.ch].set(WORD=wordsize)

    @property
    def framesize(self):
        self.regs.CHAN_BUSWIDTH[self.ch].reg.FRAME

    @framesize.setter
    def framesize(self, framesize):
        return self.regs.CHAN_BUSWIDTH[self.ch].set(FRAME=framesize)

    @property
    def burstsize(self):
        return self.regs.CHAN_BURSTSIZE[self.ch].val

    @burstsize.setter
    def burstsize(self, size):
        self.regs.CHAN_BURSTSIZE[self.ch].val = size

    @property
    def sram_carveout(self):
        reg = self.regs.CHAN_SRAM_CARVEOUT[self.ch].reg
        return (reg.BASE, reg.SIZE)

    @sram_carveout.setter
    def sram_carveout(self, carveout):
        base, size = carveout
        self.regs.CHAN_SRAM_CARVEOUT[self.ch].reg = \
                    R_CARVEOUT(BASE=base, SIZE=size)

    @property
    def DESC_WRITE(self):
        if self.tx:
            return self.regs.TX_DESC_WRITE[self.ch//2]
        else:
            return self.regs.RX_DESC_WRITE[self.ch//2]

    @property
    def REPORT_READ(self):
        if self.tx:
            return self.regs.TX_REPORT_READ[self.ch//2]
        else:
            return self.regs.RX_REPORT_READ[self.ch//2]

    def can_submit(self):
        return not self.regs.CHAN_DESC_RING[self.ch].reg.FULL

    def submit_desc(self, desc):
        if self.regs.CHAN_DESC_RING[self.ch].reg.FULL:
            raise Exception(f"ch{self.ch} descriptor ring full")

        if self.p.debug:
            print(f"admac: submitting (ch{self.ch}): {desc}", file=sys.stderr)

        for piece in desc.ser():
            self.DESC_WRITE.val = piece

        self._submitted[desc.flags.DESC_ID] = desc

    def submit(self, data=None, buflen=None, **kwargs):
        if self.tx:
            assert data is not None
            buflen = len(data)
        else:
            assert buflen is not None

        iova = self.p.get_buffer(buflen)
        if self.tx:
            self.p.iowrite(iova, data)
        self.submit_desc(ADMACDescriptor(
            iova, buflen, DESC_ID=self._desc_id, NOTIFY=1, **kwargs
        ))
        self._desc_id = (self._desc_id + 1) % 256

    def read_reports(self):
        data = bytearray()

        while not self.regs.CHAN_REPORT_RING[self.ch].reg.EMPTY:
            pieces = []
            for _ in range(4):
                pieces.append(self.REPORT_READ.val)
            report = ADMACReport.deser(pieces)

            if report.flags.DESC_ID in self._submitted:
                desc = self._submitted[report.flags.DESC_ID]
            else:
                print(f"admac: stray report (ch{self.ch}): {report}", file=sys.stderr)
                desc = None

            if self.rx and desc and self.p.dart:
                data.extend(self.p.ioread(desc.addr, desc.length))

            if self.p.debug:
                if self._last_report and desc:
                    countval_delta = report.countval - self._last_report.countval
                    est_rate = 24e6*desc.length/countval_delta/4
                    est = f"(estimated rate: {est_rate:.2f} dwords/s)"
                else:
                    est = ""

                print(f"admac: picked up (ch{self.ch}): {report} {est}", file=sys.stderr)

            self._last_report = report

        return data if self.rx else None

    @property
    def status(self):
        return self.regs.CHAN_STATUS[self.ch, 0].reg

    def poll(self, wait=True):
        while not (self.status.DESC_DONE or self.status.RING_ERR):
            time.sleep(0.001)

            if not wait:
                break

        self.regs.CHAN_STATUS[self.ch,0].reg = R_CHAN_STATUS(DESC_DONE=1)

        if self.status.RING_ERR:
            if self.p.debug:
                print(f"STATUS={self.regs.CHAN_STATUS[self.ch,1].reg} " + \
                      f"REPORT_RING={self.regs.CHAN_DESC_RING[self.ch]} " + \
                      f"DESC_RING={self.regs.CHAN_REPORT_RING[self.ch]}",
                      file=sys.stderr)
            self.regs.CHAN_DESC_RING[self.ch].set(ERR=1)
            self.regs.CHAN_REPORT_RING[self.ch].set(ERR=1)

        return self.read_reports()


class ADMAC(Reloadable):
    def __init__(self, u, devpath, dart=None, dart_stream=2,
                 reserved_size=4*1024*1024, debug=False):
        self.u = u
        self.p = u.proxy
        self.debug = debug

        if type(devpath) is str:
            adt_node = u.adt[devpath]
            # ADT's #dma-channels counts pairs of RX/TX channel, so multiply by two
            self.nchans = adt_node._properties["#dma-channels"] * 2
            self.base, _ = adt_node.get_reg(0)
        else:
            self.base = devpath
            self.nchans = 26

        self.regs = ADMACRegs(u, self.base)
        self.dart, self.dart_stream = dart, dart_stream

        if dart is not None:
            resmem_phys = u.heap.memalign(128*1024, reserved_size)
            self.resmem_iova = self.dart.iomap(dart_stream, resmem_phys, reserved_size)
            self.resmem_size = reserved_size
            self.resmem_pos = 0
            self.dart.invalidate_streams(1 << dart_stream)

        self.chans = [ADMACChannel(self, no) for no in range(self.nchans)]

    def ioread(self, base, size):
        assert self.dart is not None
        return self.dart.ioread(self.dart_stream, base, size)

    def iowrite(self, base, data):
        assert self.dart is not None
        self.dart.iowrite(self.dart_stream, base, data)

    def fill_canary(self):
        ranges = self.dart.iotranslate(self.dart_stream, 
                                self.resmem_iova, self.resmem_size)
        assert len(ranges) == 1
        start, size = ranges[0]
        self.p.memset8(start, 0xba, size)

    def get_buffer(self, size):
        assert size < self.resmem_size

        if self.resmem_pos + size > self.resmem_size:
            self.resmem_pos = 0

        bufptr = self.resmem_iova + self.resmem_pos
        self.resmem_pos += size
        return bufptr



# SPDX-License-Identifier: MIT

from m1n1.utils import *
from m1n1.constructutils import Ver
from m1n1.fw.dcp.ipc import *

class DCPAVPropHandler:
    def __init__(self):
        self.dcpav_prop = {}

    def setDCPAVPropStart(self, length):
        # print(f"setDCPAVPropStart({length:#x})")
        self.dcpav_prop_len = length - 1 # off by one?
        self.dcpav_prop_off = 0
        self.dcpav_prop_data = []
        return True

    def setDCPAVPropChunk(self, data, offset, length):
        # print(f"setDCPAVPropChunk(..., {offset:#x}, {length:#x})")
        assert offset == self.dcpav_prop_off
        self.dcpav_prop_data.append(data)
        self.dcpav_prop_off += len(data)
        return True

    def setDCPAVPropEnd(self, key):
        # print(f"setDCPAVPropEnd({key!r})")
        blob = b"".join(self.dcpav_prop_data)
        assert self.dcpav_prop_len == len(blob)
        self.dcpav_prop[key] = OSSerialize().parse(blob)
        self.dcpav_prop_data = self.dcpav_prop_len = self.dcpav_prop_off = None
        pprint.pprint(self.dcpav_prop[key])
        return True


def parse_log(fd):
    op_stack = {}
    for line in fd:
        optype, args = line.split(" ", 1)
        if optype == "CALL":
            d, msg, chan, off, msg, in_size, out_size, in_data = args.split(" ")
            op = Call(d, chan, int(off, 0), msg, int(in_size, 0), int(out_size, 0),
                      bytes.fromhex(in_data))
            op_stack.setdefault(chan, []).append(op)
        elif optype == "ACK":
            d, msg, chan, off, out_data = args.split(" ")
            op = op_stack[chan].pop()
            assert int(off, 0) == op.off
            op.ack(bytes.fromhex(out_data))
        elif optype == "VERSION":
            for arg in args.strip().split(" "):
                version = arg.split(":")
                if len(version) == 2:
                    Ver.set_version_key(version[0], version[1])
            continue
        else:
            raise Exception(f"Unknown log cmd {optype}")

        yield op

def dump_log(fd):
    nesting = {
        "": 0,
        "OOB": 0,
    }

    handler = DCPAVPropHandler()

    for op in parse_log(fd):
        ctx = ""
        if Ver.check("V < V13_5"):
            dcpavprop_cbs = ["D122", "D123", "D124"]
        else:
             dcpavprop_cbs = ["D126", "D127", "D128"]
        if not op.complete and op.msg in dcpavprop_cbs:
            method = op.get_method()
            if op.msg == dcpavprop_cbs[0]:
                method.callback(handler.setDCPAVPropStart, op.in_data)
            if op.msg == dcpavprop_cbs[1]:
                method.callback(handler.setDCPAVPropChunk, op.in_data)
            if op.msg == dcpavprop_cbs[2]:
                method.callback(handler.setDCPAVPropEnd, op.in_data)

        if "OOB" in op.chan:
            ctx = "[OOB] -----------> "
        if not op.complete:
            op.print_req(indent=ctx + "  " * nesting.setdefault(ctx, 0))
            nesting[ctx] += 1
        else:
            nesting[ctx] -= 1
            op.print_reply(indent=ctx + "  " * nesting.setdefault(ctx, 0))

if __name__ == "__main__":
    import sys
    dump_log(open(sys.argv[1]))



# SPDX-License-Identifier: MIT

from ..hv import TraceMode
from ..utils import *

__all__ = []

class RegCacheAlwaysCached(Reloadable):
    def __init__(self, parent):
        self.parent = parent

    def read(self, addr, width):
        return self.parent.read_cached(addr, width)

    def write(self, addr, data, width):
        raise Exception("Trying to write a register to the cache")

class RegCache(Reloadable):
    def __init__(self, hv):
        self.hv = hv
        self.u = hv.u
        self.cache = {}

        self.cached = RegCacheAlwaysCached(self)

    def update(self, addr, data):
        self.cache[addr] = data

    def read(self, addr, width):
        if self.hv.ctx or not self.hv.started:
            data = self.u.read(addr, width)
            self.cache[addr] = data
            return data
        else:
            return self.read_cached(addr, width)

    def read_cached(self, addr, width):
        data = self.cache.get(addr, None)
        if data is None:
            print(f"RegCache: no cache for {addr:#x}")
        return data

    def write(self, addr, data, width):
        if self.hv.ctx:
            self.u.write(addr, data, width)
            self.cache[addr] = data
        else:
            raise Exception("Cannot write register in asynchronous context")

class TracerState:
    pass

class Tracer(Reloadable):
    DEFAULT_MODE = TraceMode.ASYNC

    def __init__(self, hv, verbose=False, ident=None):
        self.hv = hv
        self.ident = ident or type(self).__name__
        self.regmaps = {}
        self.verbose = verbose
        self.state = TracerState()
        self.init_state()
        self._cache = RegCache(hv)
        cache = hv.tracer_caches.get(self.ident, None)
        if cache is not None:
            self._cache.cache.update(cache.get("regcache", {}))
            self.state.__dict__.update(cache.get("state", {}))
        hv.tracer_caches[self.ident] = {
            "regcache": self._cache.cache,
            "state": self.state.__dict__
        }

    def init_state(self):
        pass

    def hook_w(self, addr, val, width, **kwargs):
        self.hv.u.write(addr, val, width)

    def hook_r(self, addr, width, **kwargs):
        return self.hv.u.read(addr, width)

    def evt_rw(self, evt, regmap=None, prefix=None):
        self._cache.update(evt.addr, evt.data)
        reg = rcls = None
        value = evt.data

        t = "w" if evt.flags.WRITE else "r"

        if regmap is not None:
            reg, index, rcls = regmap.lookup_addr(evt.addr)
            if rcls is not None:
                value = rcls(evt.data)

        if self.verbose >= 3 or (reg is None and self.verbose >= 1):
            if reg is None:
                s = f"{evt.addr:#x} = {value:#x}"
            else:
                s = f"{regmap.get_name(evt.addr)} = {value!s}"
            m = "+" if evt.flags.MULTI else " "
            self.log(f"MMIO: {t.upper()}.{1<<evt.flags.WIDTH:<2}{m} " + s)

        if reg is not None:
            if prefix is not None:
                attr = f"{t}_{prefix}_{reg}"
            else:
                attr = f"{t}_{reg}"
            handler = getattr(self, attr, None)
            if handler:
                if index is not None:
                    handler(value, index)
                else:
                    handler(value)
            elif self.verbose == 2:
                s = f"{regmap.get_name(evt.addr)} = {value!s}"
                m = "+" if evt.flags.MULTI else " "
                self.log(f"MMIO: {t.upper()}.{1<<evt.flags.WIDTH:<2}{m} " + s)

    def trace(self, start, size, mode, read=True, write=True, **kwargs):
        zone = irange(start, size)
        if mode == TraceMode.HOOK:
            self.hv.add_tracer(zone, self.ident, mode, self.hook_r if read else None,
                               self.hook_w if write else None, **kwargs)
        else:
            self.hv.add_tracer(zone, self.ident, mode, self.evt_rw if read else None,
                               self.evt_rw if write else None, **kwargs)

    def trace_regmap(self, start, size, cls, mode=None, name=None, prefix=None, regmap_offset=0):
        if mode is None:
            mode = self.DEFAULT_MODE
        if name is None:
            name = cls.__name__

        regmap = self.regmaps.get(start - regmap_offset, None)
        if regmap is None:
            regmap = cls(self._cache, start - regmap_offset)
            regmap.cached = cls(self._cache.cached, start - regmap_offset)
            self.regmaps[start - regmap_offset] = regmap
        else:
            assert isinstance(regmap, cls)

        setattr(self, name, regmap)
        self.trace(start, size, mode=mode, regmap=regmap, prefix=prefix)

    def start(self):
        pass

    def stop(self):
        self.hv.clear_tracers(self.ident)

    def log(self, msg, show_cpu=True):
        self.hv.log(f"[{self.ident}] {msg}", show_cpu=show_cpu)

class PrintTracer(Tracer):
    def __init__(self, hv, device_addr_tbl):
        super().__init__(hv)
        self.device_addr_tbl = device_addr_tbl
        self.log_file = None

    def event_mmio(self, evt, name=None, start=None):
        dev, zone2 = self.device_addr_tbl.lookup(evt.addr)
        if name is None:
            name = dev
            start = zone2.start
        t = "W" if evt.flags.WRITE else "R"
        m = "+" if evt.flags.MULTI else " "
        logline = (f"[cpu{evt.flags.CPU}] [0x{evt.pc:016x}] MMIO: {t}.{1<<evt.flags.WIDTH:<2}{m} " +
                   f"0x{evt.addr:x} ({name}, offset {evt.addr - start:#04x}) = 0x{evt.data:x}")
        print(logline)
        if self.log_file:
            self.log_file.write(f"# {logline}\n")
            width = 8 << evt.flags.WIDTH
            if evt.flags.WRITE:
                stmt = f"p.write{width}({start:#x} + {evt.addr - start:#x}, {evt.data:#x})\n"
            else:
                stmt = f"p.read{width}({start:#x} + {evt.addr - start:#x})\n"
            self.log_file.write(stmt)

class ADTDevTracer(Tracer):
    REGMAPS = []
    NAMES = []
    PREFIXES = []

    def __init__(self, hv, devpath, verbose=False):
        super().__init__(hv, verbose=verbose, ident=type(self).__name__ + "@" + devpath)
        self.dev = hv.adt[devpath]

    @classmethod
    def _reloadcls(cls, force=False):
        regmaps = []
        for i in cls.REGMAPS:
            if i is None:
                reloaded = None
            elif isinstance(i, tuple):
                reloaded = (i[0]._reloadcls(force), i[1])
            else:
                reloaded = i._reloadcls(force)
            regmaps.append(reloaded)
        cls.REGMAPS = regmaps

        return super()._reloadcls(force)

    def start(self):
        for i in range(len(self.dev.reg)):
            if i >= len(self.REGMAPS) or (regmap := self.REGMAPS[i]) is None:
                continue
            if isinstance(regmap, tuple):
                regmap, regmap_offset = regmap
            else:
                regmap_offset = 0
            prefix = name = None
            if i < len(self.NAMES):
                name = self.NAMES[i]
            if i < len(self.PREFIXES):
                prefix = self.PREFIXES[i]

            start, size = self.dev.get_reg(i)
            self.trace_regmap(start, size, regmap, name=name, prefix=prefix, regmap_offset=regmap_offset)

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__.startswith(__name__))



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1 import asm

code_len = 12 * 16 * 8 + 4
data_len = 8 * 16 * 8

if u.mrs(SPRR_CONFIG_EL1):
    u.msr(GXF_CONFIG_EL12, 0)
    u.msr(SPRR_CONFIG_EL12, 0)
    u.msr(GXF_CONFIG_EL1, 0)
    u.msr(SPRR_CONFIG_EL1, 0)

u.msr(HACR_EL2, 0)

hcr = HCR(u.mrs(HCR_EL2))
hcr.TIDCP = 0
hcr.TGE = 0
u.msr(HCR_EL2, hcr.value)
u.inst(0xd5033fdf) # isb

ACTLR_DEFAULT = 0xc00
ACTLR_AFP = 1 << 5
u.msr(ACTLR_EL1, ACTLR_DEFAULT | ACTLR_AFP)

code_buffer = p.malloc(code_len)
data_buffer = p.malloc(data_len)

template = asm.ARMAsm("""
    mov x2, x0
    mrs x2, s3_0_c0_c0_0
    str x2, [x1], #8
    ret
""", code_buffer)

mov, mrs, st, ret = struct.unpack("4I", template.data)

data = []

BAD = 0xacce5515abad1dea

AUX = [
    ACTLR_EL1,
    ACTLR_EL2,
    AFSR0_EL1,
    AFSR0_EL2,
    AFSR1_EL1,
    AFSR1_EL2,
    AIDR_EL1,
    AIDR2_EL1,
    AMAIR_EL1,
    AMAIR_EL2,
    APCTL_EL1,
    APSTS_EL1,
]

def test():
    u.msr(SPRR_CONFIG_EL1, 1)
    u.msr(GXF_CONFIG_EL1, 1)
    u.msr(SPRR_CONFIG_EL12, 1)
    u.msr(GXF_CONFIG_EL12, 1)

    for op1 in range(1 << 3):
        for CRn in (0b1011, 0b1111):
            mrs0 = mrs | (op1 << 16) | (CRn << 12)
            insns = []
            for CRm in range(1 << 4):
                for op2 in range(1 << 3):
                    insns.extend((mov, mrs0 | (CRm << 8) | (op2 << 5), st))
            insns.append(ret)
            iface.writemem(code_buffer, struct.pack("<385I", *insns))
            p.dc_cvau(code_buffer, code_len)
            p.ic_ivau(code_buffer, code_len)

            p.set_exc_guard(GUARD.SILENT | GUARD.SKIP)
            p.el1_call(code_buffer, BAD, data_buffer)
            cnt = p.get_exc_count()

            data = iface.readmem(data_buffer, data_len)
            d = struct.unpack("<128Q", data)
            i = 0
            for CRm in range(1 << 4):
                for op2 in range(1 << 3):
                    v = d[i]
                    if v != BAD:
                        yield (3, op1, CRn, CRm, op2)
                    i += 1
    for enc in AUX:
        try:
            v = u.mrs(enc, call="el1", silent=True)
            if v != BAD:
                yield enc
        except:
            continue

    u.msr(GXF_CONFIG_EL12, 0)
    u.msr(SPRR_CONFIG_EL12, 0)
    u.msr(GXF_CONFIG_EL1, 0)
    u.msr(SPRR_CONFIG_EL1, 0)

baseline = set(test())

for bit in range(64):
    print()
    print ("## HACR_EL2[%d]" % bit)
    u.msr(HACR_EL2, 1<<bit)
    u.inst(0xd5033fdf) # isb

    new = set(test())

    added = new - baseline
    removed = baseline - new

    if added:
        print("Untraps:")
        for enc in sorted(added):
            print(f"{sysreg_name(enc)} ({', '.join(str(i) for i in enc)})")

    if removed:
        print("Traps:")
        for enc in sorted(removed):
            print(f"{sysreg_name(enc)} ({', '.join(str(i) for i in enc)})")

p.set_exc_guard(GUARD.OFF)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct
from m1n1.setup import *
from m1n1.fw.smc import SMCClient, SMCError

smc_addr = u.adt["arm-io/smc"].get_reg(0)[0]
smc = SMCClient(u, smc_addr)
smc.start()
smc.start_ep(0x20)

smc.verbose = 0

smcep = smc.epmap[0x20]

def gpio_key(pin):
    assert(pin < (1 << 16))

    fourcc = 'gP' + ('00'+(hex(pin)[2:]))[-2:]
    return fourcc

## Enable wifi/bluetooth
#RFKILL_PIN = 13
#smcep.write(gpio_key(RFKILL_PIN), struct.pack('<I', 0x800000 | 0x0))
#smcep.write(gpio_key(RFKILL_PIN), struct.pack('<I', 0x800000 | 0x1))

count = smcep.read32b("#KEY")
print(f"Key count: {count}")

for i in range(count):
    k = smcep.get_key_by_index(i)
    length, type, flags = smcep.get_key_info(k)
    if flags & 0x80:
        try:
            val = smcep.read_type(k, length, type)
            print(f"#{i}: {k} = ({type}, {flags:#x}) {val}")
        except SMCError as e:
            print(f"#{i}: {k} = ({type}, {flags:#x}) <error {e}>")
    else:
        print(f"#{i}: {k} = ({type}, {flags:#x}) <not available>")


smc.stop()



# SPDX-License-Identifier: MIT
import json, os, re
from enum import Enum, IntEnum, IntFlag
from .utils import Register, Register64, Register32

__all__ = ["sysreg_fwd", "sysreg_rev"]

def _load_registers():
    global sysreg_fwd, sysop_fwd

    sysreg_fwd = {}
    sysop_fwd = {}
    for fname in ["arm_regs.json", "apple_regs.json"]:
        data = json.load(open(os.path.join(os.path.dirname(__file__), "..", "..", "tools", fname)))
        for reg in data:
            if "accessors" in reg:
                for acc in reg["accessors"]:
                    if acc in ("MRS", "MSR"):
                        sysreg_fwd[reg["name"]] = tuple(reg["enc"])
                    else:
                        sysop_fwd[acc + " " + reg["name"]] = tuple(reg["enc"])
            else:
                sysreg_fwd[reg["name"]] = tuple(reg["enc"])

_load_registers()
sysreg_rev = {v: k for k, v in sysreg_fwd.items()}
sysop_rev = {v: k for k, v in sysop_fwd.items()}
sysop_fwd_id = {k.replace(" ", "_"): v for k,v in sysop_fwd.items()}

globals().update(sysreg_fwd)
__all__.extend(sysreg_fwd.keys())
globals().update(sysop_fwd_id)
__all__.extend(sysop_fwd_id.keys())

def sysreg_name(enc):
    if enc in sysreg_rev:
        return sysreg_rev[enc]
    if enc in sysop_rev:
        return sysop_rev[enc]
    return f"s{enc[0]}_{enc[1]}_c{enc[2]}_c{enc[3]}_{enc[4]}"

def sysreg_parse(s):
    if isinstance(s, tuple) or isinstance(s, list):
        return tuple(s)
    s = s.strip()
    for r in (r"s(\d+)_(\d+)_c(\d+)_c(\d+)_(\d+)", r"(\d+), *(\d+), *(\d+), *(\d+), *(\d+)"):
        if m := re.match(r, s):
            enc = tuple(map(int, m.groups()))
            break
    else:
        for i in sysreg_fwd, sysop_fwd, sysop_fwd_id:
            try:
                enc = i[s]
            except KeyError:
                continue
            break
        else:
            raise Exception(f"Unknown sysreg name {s}")
    return enc

def DBGBCRn_EL1(n):
    return (2,0,0,n,5)

def DBGBVRn_EL1(n):
    return (2,0,0,n,4)

def DBGWCRn_EL1(n):
    return (2,0,0,n,7)

def DBGWVRn_EL1(n):
    return (2,0,0,n,6)

class ESR_EC(IntEnum):
    UNKNOWN        = 0b000000
    WFI            = 0b000001
    FP_TRAP        = 0b000111
    PAUTH_TRAP     = 0b001000
    LS64           = 0b001010
    BTI            = 0b001101
    ILLEGAL        = 0b001110
    SVC            = 0b010101
    HVC            = 0b010110
    SMC            = 0b010111
    MSR            = 0b011000
    SVE            = 0b011001
    PAUTH_FAIL     = 0b011100
    IABORT_LOWER   = 0b100000
    IABORT         = 0b100001
    PC_ALIGN       = 0b100010
    DABORT_LOWER   = 0b100100
    DABORT         = 0b100101
    SP_ALIGN       = 0b100110
    FP_EXC         = 0b101100
    SERROR         = 0b101111
    BKPT_LOWER     = 0b110000
    BKPT           = 0b110001
    SSTEP_LOWER    = 0b110010
    SSTEP          = 0b110011
    WATCH_LOWER    = 0b110100
    WATCH          = 0b110101
    BRK            = 0b111100
    IMPDEF         = 0b111111

class MSR_DIR(IntEnum):
    WRITE = 0
    READ = 1

class ESR_ISS_MSR(Register32):
    Op0 = 21, 20
    Op2 = 19, 17
    Op1 = 16, 14
    CRn = 13, 10
    Rt = 9, 5
    CRm = 4, 1
    DIR = 0, 0, MSR_DIR

class DABORT_DFSC(IntEnum):
    ASIZE_L0         = 0b000000
    ASIZE_L1         = 0b000001
    ASIZE_L2         = 0b000010
    ASIZE_L3         = 0b000011
    XLAT_L0          = 0b000100
    XLAT_L1          = 0b000101
    XLAT_L2          = 0b000110
    XLAT_L3          = 0b000111
    AF_L0            = 0b001000
    AF_L1            = 0b001001
    AF_L2            = 0b001010
    AF_L3            = 0b001011
    PERM_L0          = 0b001100
    PERM_L1          = 0b001101
    PERM_L2          = 0b001110
    PERM_L3          = 0b001111
    EABORT           = 0b010000
    TAG_CHECK        = 0b010001
    PT_EABORT_Lm1    = 0b010011
    PT_EABORT_L0     = 0b010100
    PT_EABORT_L1     = 0b010101
    PT_EABORT_L2     = 0b010110
    PT_EABORT_L3     = 0b010111
    ECC_ERROR        = 0b011000
    PT_ECC_ERROR_Lm1 = 0b011011
    PT_ECC_ERROR_L0  = 0b011100
    PT_ECC_ERROR_L1  = 0b011101
    PT_ECC_ERROR_L2  = 0b011110
    PT_ECC_ERROR_L3  = 0b011111
    ALIGN            = 0b100001
    ASIZE_Lm1        = 0b101001
    XLAT_Lm1         = 0b101011
    TLB_CONFLICT     = 0b110000
    UNSUPP_ATOMIC    = 0b110001
    IMPDEF_LOCKDOWN  = 0b110100
    IMPDEF_ATOMIC    = 0b110101

class ESR_ISS_DABORT(Register32):
    ISV = 24
    SAS = 23, 22
    SSE = 21
    SRT = 20, 16
    SF = 15
    AR = 14
    VNCR = 13
    SET = 12, 11
    LSR = 12, 11
    FnV = 10
    EA = 9
    CM = 8
    S1PTR = 7
    WnR = 6
    DFSC = 5, 0, DABORT_DFSC

class ESR(Register64):
    ISS2 = 36, 32
    EC = 31, 26, ESR_EC
    IL = 25
    ISS = 24, 0

class SPSR_M(IntEnum):
    EL0t = 0
    EL1t = 4
    EL1h = 5
    EL2t = 8
    EL2h = 9

class SPSR(Register64):
    N = 31
    Z = 30
    C = 29
    V = 28
    TCO = 25
    DIT = 24
    UAO = 23
    PAN = 22
    SS = 21
    IL = 20
    SSBS = 12
    BTYPE = 11, 10
    D = 9
    A = 8
    I = 7
    F = 6
    M = 4, 0, SPSR_M

class ACTLR(Register64):
    EnMDSB  = 12
    EnPRSV  = 6
    EnAFP   = 5
    EnAPFLG = 4
    DisHWP  = 3
    EnTSO   = 1

class HCR(Register64):
    TWEDEL   = 63, 60
    TWEDEn   = 59
    TID5     = 58
    DCT      = 57
    ATA      = 56
    TTLBOS   = 55
    TTLBIS   = 54
    EnSCXT   = 53
    TOCU     = 52
    AMVOFFEN = 51
    TICAB    = 50
    TID4     = 49
    FIEN     = 47
    FWB      = 46
    NV2      = 45
    AT       = 44
    NV1      = 43
    NV       = 42
    API      = 41
    APK      = 40
    MIOCNCE  = 38
    TEA      = 37
    TERR     = 36
    TLOR     = 35
    E2H      = 34
    ID       = 33
    CD       = 32
    RW       = 31
    TRVM     = 30
    HCD      = 29
    TDZ      = 28
    TGE      = 27
    TVM      = 26
    TTLB     = 25
    TPU      = 24
    TPCP     = 23
    TPC      = 23
    TSW      = 22
    TACR     = 21
    TIDCP    = 20
    TSC      = 19
    TID3     = 18
    TID2     = 17
    TID1     = 16
    TID0     = 15
    TWE      = 14
    TWI      = 13
    DC       = 12
    BSU      = 11, 10
    FB       = 9
    VSE      = 8
    VI       = 7
    VF       = 6
    AMO      = 5
    IMO      = 4
    FMO      = 3
    PTW      = 2
    SWIO     = 1
    VM       = 0

class HACR(Register64):
    TRAP_CPU_EXT = 0
    TRAP_AIDR = 4
    TRAP_AMX = 10
    TRAP_SPRR = 11
    TRAP_GXF = 13
    TRAP_CTRR = 14
    TRAP_IPI = 16
    TRAP_s3_4_c15_c5z6_x = 18
    TRAP_s3_4_c15_c0z12_5 = 19
    GIC_CNTV = 20
    TRAP_s3_4_c15_c10_4 = 25
    TRAP_SERROR_INFO = 48
    TRAP_EHID = 49
    TRAP_HID = 50
    TRAP_s3_0_c15_c12_1z2 = 51
    TRAP_ACC = 52
    TRAP_PM = 57
    TRAP_UPM = 58
    TRAP_s3_1z7_c15_cx_3 = 59

class AMX_CONFIG(Register64):
    EN = 63
    EN_EL1 = 62

class MDCR(Register64):
    TDE = 8
    TDA = 9
    TDOSA = 10
    TDRA = 11

class MDSCR(Register64):
    SS = 0
    MDE = 15

class DBGBCR(Register32):
    BT = 23, 20
    LBN = 16, 16
    SSC = 15, 14
    HMC = 13
    BAS = 8,5
    PMC = 2,1
    E = 0

class DBGWCR_LSC(IntFlag):
    L = 1
    S = 2

class DBGWCR(Register32):
    SSCE = 29
    MASK = 28, 24
    WT = 20
    LBN = 19, 16
    SSC = 15, 14
    HMC = 13
    BAS = 12, 5
    LSC = 4, 3
    PAC = 2, 1
    E = 0

# TCR_EL1
class TCR(Register64):
    DS = 59
    TCMA1 = 58
    TCMA0 = 57
    E0PD1 = 56
    E0PD0 = 55
    NFD1 = 54
    NFD0 = 53
    TBID1 = 52
    TBID0 = 51
    HWU162 = 50
    HWU161 = 49
    HWU160 = 48
    HWU159 = 47
    HWU062 = 46
    HWU061 = 45
    HWU060 = 44
    HWU059 = 43
    HPD1 = 42
    HPD0 = 41
    HD = 40
    HA = 39
    TBI1 = 38
    TBI0 = 37
    AS = 36
    IPS = 34, 32
    TG1 = 31, 30
    SH1 = 29, 28
    ORGN1 = 27, 26
    IRGN1 = 25, 24
    EPD1 = 23
    A1 = 22
    T1SZ = 21, 16
    TG0 = 15, 14
    SH0 = 13, 12
    ORGN0 = 11, 10
    IRGN0 = 9, 8
    EPD0 = 7
    T0SZ = 5, 0

class TLBI_RVA(Register64):
    ASID = 63, 48
    TG = 47, 46
    SCALE = 45, 44
    NUM = 43, 39
    TTL = 38, 37
    BaseADDR = 36, 0

class MIDR_PART(IntEnum):
    T8110_BLIZZARD = 0x30

class MIDR(Register64):
    REV_LOW = 3, 0
    PART = 15, 4
    REV_HIGH = 23, 20

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)



# SPDX-License-Identifier: MIT
from ..utils import *
from enum import IntEnum

class R_STATUS(Register32):
    EN  = 0
    RST = 1

class R_MCLK_CONF(Register32):
    SEL = 3, 0

class R_PORT_ENABLES(Register32):
    CLOCK1 = 1
    CLOCK2 = 2
    DATA   = 3

class R_PORT_CLKSEL(Register32):
    SEL = 11, 8

class R_PORT_DATASEL(Register32):
    TXA0 = 0
    TXA1 = 2
    TXA2 = 4
    TXA3 = 6
    TXA4 = 8
    TXA5 = 10

    TXB0 = 1
    TXB1 = 3
    TXB2 = 5
    TXB3 = 7
    TXB4 = 9
    TXB5 = 11

class E_SLOT_WIDTH(IntEnum):
    NONE = 0

    W_16BIT =  0x4
    W_20BIT =  0x8
    W_24BIT =  0xc
    W_32BIT = 0x10

class R_SERDES_CONF(Register32):
    NSLOTS     = 3, 0
    SLOT_WIDTH = 8, 4, E_SLOT_WIDTH

    BCLK_POL  = 10
    LSB_FIRST = 11

    UNK1 = 12
    UNK2 = 13
    IDLE_UNDRIVEN = 14 # TX only
    NO_DATA_FEEDBACK = 15 # RX only

    SYNC_SEL = 18, 16

class R_INTMASK(Register32):
    # macOS interested in 0x823c
    UNK1 = 2 # m
    UNK2 = 3 # m
    UNK3 = 4 # m
    TX_UNDERFLOW = 5 # m

    UNK4 = 9 # m
    READ_SENSITIVE_UNK1 = 11
    READ_SENSITIVE_UNK2 = 15 # m

class MCAClusterRegs(RegMap):
    MCLK_STATUS = 0x0, R_STATUS
    MCLK_CONF   = 0x4, R_MCLK_CONF

    SYNCGEN_STATUS    = 0x100, R_STATUS
    SYNCGEN_MCLK_SEL  = 0x104, Register32
    SYNCGEN_HI_PERIOD = 0x108, Register32
    SYNCGEN_LO_PERIOD = 0x10c, Register32

    PORT_ENABLES   = 0x600, R_PORT_ENABLES
    PORT_CLK_SEL   = 0x604, R_PORT_CLKSEL
    PORT_DATA_SEL  = 0x608, R_PORT_DATASEL

    INTSTATE = 0x700, R_INTMASK
    INTMASK  = 0x704, R_INTMASK

class MCATXSerdesRegs(RegMap):
    STATUS   = 0x0, R_STATUS
    CONF     = 0x4, R_SERDES_CONF
    BITDELAY = 0x8, Register32
    CHANMASK = irange(0xc, 4, 4), Register32

class MCARXSerdesRegs(RegMap):
    STATUS   = 0x0, R_STATUS
    UNK1     = 0x4, Register32
    CONF     = 0x8, R_SERDES_CONF
    BITDELAY = 0xc, Register32
    CHANMASK = irange(0x10, 4, 4), Register32


class MCACluster:
    def __init__(self, u, base):
        self.regs = MCAClusterRegs(u, base)
        self.txa = MCATXSerdesRegs(u, base + 0x300)
        self.txb = MCATXSerdesRegs(u, base + 0x500)
        self.rxa = MCARXSerdesRegs(u, base + 0x200)
        self.rxb = MCARXSerdesRegs(u, base + 0x400)
        self.all_regs = [
            self.regs,
            self.txa, self.txb,
            self.rxa, self.rxb
        ]




# SPDX-License-Identifier: MIT
from construct import *

from ...utils import *
from ..asc import StandardASC
from ..afk.epic import *

class DCPAVControllerService(EPICStandardService):
    NAME = "dcpav-controller-epic"
    SHORT = "dcpav"

    def setPower(self, power):
        self.call(8, 0x8, struct.pack("<16xI12x", power))

    def getPower(self, power):
        return struct.unpack("<16xI12x", self.call(8, 0x9, bytes(32)))

    def wakeDisplay(self):
        self.call(8, 0xa, bytes(16))

    def sleepDisplay(self):
        self.call(8, 0xb, bytes(16))

    def forceHotPlugDetect(self):
        self.call(8, 0xc, bytes(16))

    def setVirtualDeviceMode(self, mode):
        self.call(8, 0xd, struct.pack("<16xI12x", mode))

class DCPDPControllerService(EPICStandardService):
    NAME = "dcpdp-controller-epic"
    SHORT = "dcpdp"

class DCPDPTXEndpoint(EPICEndpoint):
    SHORT = "dptx"

    SERVICES = [
        DCPAVControllerService,
        DCPDPControllerService,
    ]

ATC0 = 0
ATC1 = 1
ATC2 = 2
ATC3 = 3
LPDPTX = 4
DPTX = 5

DPPHY = 0
DPIN0 = 1
DPIN1 = 2

class DCPDPTXRemotePortService(EPICStandardService):
    NAME = "dcpdptx-port-epic"
    SHORT = "port"

    def displayRequest(self):
        self.call(8, 8, bytes(16))

    def displayRelease(self):
        self.call(8, 9, bytes(16))

    def connectTo(self, connected, unit, port, unk=0):
        target = 0
        if connected:
            target |= (1 << 8)
        target |= unit
        target |= port << 4
        self.call(8, 13, struct.pack("<16xII8x", unk, target))

class DCPDPTXPortEndpoint(EPICEndpoint):
    SHORT = "dpport"

    SERVICES = [
        DCPDPTXRemotePortService,
        DCPDPControllerService,
    ]

class DCPDPDevice(EPICStandardService):
    NAME = "dcpav-device-epic"
    SHORT = "dpdev"

class DCPAVDeviceEndpoint(EPICEndpoint):
    SHORT = "avdev"

    SERVICES = [
        DCPDPDevice,
    ]

class DCPDPService(EPICStandardService):
    NAME = "dcpav-service-epic"
    SHORT = "dpserv"

class DCPAVServiceEndpoint(EPICEndpoint):
    SHORT = "avserv"

    SERVICES = [
        DCPDPService,
    ]

class DCPAVSimpleVideoInterface(EPICStandardService):
    NAME = "dcpav-video-interface-epic"
    SHORT = "video"

class DCPAVVideoEndpoint(EPICEndpoint):
    SHORT = "avserv"

    SERVICES = [
        DCPAVSimpleVideoInterface,
    ]



# SPDX-License-Identifier: MIT
from .base import *
from ...utils import *

## OSLog endpoint

class OSLogMessage(Register64):
    TYPE = 63, 56

class OSLog_GetBuf(OSLogMessage):
    TYPE = 63, 56, Constant(1)
    SIZE = 55, 48
    DVA = 47, 0

class ASCOSLogEndpoint(ASCBaseEndpoint):
    BASE_MESSAGE = OSLogMessage
    SHORT = "iorep"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.iobuffer = None
        self.iobuffer_dva = None

    @msg_handler(1, OSLog_GetBuf)
    def GetBuf(self, msg):
        if self.iobuffer:
            self.log("WARNING: trying to reset iobuffer!")


        if msg.DVA != 0:
            self.bufsize = 0x1000 * msg.SIZE
            self.iobuffer = self.iobuffer_dva = msg.DVA << 12
            self.log(f"buf prealloc {self.iobuffer:#x} / {self.iobuffer_dva:#x}")
        else:
            self.bufsize = align(0x1000 * msg.SIZE, 0x4000)
            self.iobuffer, self.iobuffer_dva = self.asc.ioalloc(self.bufsize)
            self.log(f"buf {self.iobuffer:#x} / {self.iobuffer_dva:#x}")
            self.send(OSLog_GetBuf(DVA=self.iobuffer_dva >> 12, SIZE=self.bufsize // 0x1000))

        return True
